/** @file emu.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 emulation i/f ]
*/

#include "emu.h"

#include "config.h"
#include "fileio.h"

// ----------------------------------------------------------------------------
// initialize
// ----------------------------------------------------------------------------

EMU::EMU(const char *new_app_path, const char *new_ini_path) : Logging()
{
	initialized = false;

	// get module path
	memset(app_path, 0, sizeof(app_path));
	_tcscpy(app_path, new_app_path);
	memset(ini_path, 0, sizeof(ini_path));
	_tcscpy(ini_path, new_ini_path);

	vm = NULL;
	gui = NULL;
	ledbox = NULL;
	msgboard = NULL;

#ifdef _BML3MK5
	memset(parami, 0, sizeof(parami));
	memset(paramv, 0, sizeof(paramv));
#endif

#ifdef USE_DIRECT3D
	pD3Device = NULL;
	pD3Dmixsuf = NULL;
	pD3Dsource = NULL;
	pD3Dorigin = NULL;
#ifdef USE_SCREEN_ROTATE
	pD3Drotate = NULL;
#endif
	lpD3DBmp = NULL;
#endif

	EMU_INPUT();
	EMU_SCREEN();

	// open logfile
	open_logfile(ini_path);
	out_log(LOG_INFO, _T("START [%s]"), app_path);

	//
	vm_pause = vm_syspause = vm_usrpause = false;

	// d88 bank switch
	memset(d88_file, 0, sizeof(d88_file));
	for(int drv=0; drv < MAX_DRIVE; drv++) d88_file[drv].prev_bank = -2;
}

void EMU::initialize(HWND hwnd, HINSTANCE hinst)
{
	// store main window handle
	main_window_handle = hwnd;
	instance_handle = hinst;

#ifdef USE_CPU_CLOCK_LOW
	cpu_clock_low = config.cpu_clock_low;
#endif

#ifdef _BML3MK5
	// create dialogbox
	ledbox = new LedBox(hinst);
	msgboard = new MsgBoard(hwnd, this);
	msgboard->SetVisible(config.show_dlg & 2 ? true : false);

	set_parami(EMU::ParamFddType, config.fdd_type);
	set_parami(EMU::ParamIOPort, config.io_port);
#endif

	// initialize
	vm = new VM(this);
	initialize_input();
	initialize_screen();
	initialize_sound();
#ifdef USE_FD1
	initialize_disk_insert();
#endif
#ifdef USE_MEDIA
	initialize_media();
#endif
#ifdef USE_SOCKET
	initialize_socket();
#endif
//	vm->reset();
	start_sound();

	initialized = true;
}

EMU::~EMU()
{
	out_log(LOG_INFO, _T("EXIT"));
	close_logfile();

	if (msgboard) {
		delete msgboard;
	}
}

void EMU::release()
{
	PEXCEPTION_POINTERS Ex = NULL;

	end_sound();

	update_params();
	if (ledbox) {
		delete ledbox;
	}
	__try {
		if(vm) {
			delete vm;
		}
	}
	__except (Ex = GetExceptionInformation(),
			EXCEPTION_EXECUTE_HANDLER) {
		out_log(LOG_ERROR,_T("EMU::release Exception: %x"),Ex->ExceptionRecord->ExceptionCode);
	}

	release_input();
	release_screen();
	release_sound();
#ifdef USE_MEDIA
	release_media();
#endif
#ifdef USE_SOCKET
	release_socket();
#endif
}

// ----------------------------------------------------------------------------
// drive machine
// ----------------------------------------------------------------------------

int EMU::frame_interval()
{
#ifdef SUPPORT_VARIABLE_TIMING
	return (int)(1024. * 1000. / vm->frame_rate() + 0.5);

#else
	return (int)(1024. * 1000. / FRAMES_PER_SEC + 0.5);
#endif
}

int EMU::run(int split_num)
{
	update_input();
	update_timer();
#ifdef USE_FD1
	update_disk_insert();
#endif
#ifdef USE_SOCKET
	update_socket();
#endif

	// virtual machine may be driven to fill sound buffer
	int extra_frames = 0;
	update_sound(&extra_frames);

	// drive virtual machine
	if(extra_frames == 0) {
		vm->run(split_num);
		extra_frames = 1;
	}
#ifdef _BML3MK5
	if (msgboard && split_num == 0) {
		msgboard->CountDown();
	}
#endif
	return extra_frames;
}

// update status if need
void EMU::update()
{
	update_timer();
#ifdef USE_FD1
	update_disk_insert();
#endif
#ifdef USE_SOCKET
	update_socket();
#endif

	// virtual machine may be driven to fill sound buffer
	int extra_frames = 0;
	update_sound(&extra_frames);

	return;
}

void EMU::reset()
{
	// update config
	update_params();

#ifdef USE_CPU_CLOCK_LOW
	if(cpu_clock_low != config.cpu_clock_low) {
		// stop sound
		if(sound_ok && sound_started) {
			lpdsb->Stop();
			sound_started = false;
		}
		// reinitialize virtual machine
		delete vm;
		vm = new VM(this);
		vm->initialize_sound(sound_rate, sound_samples);
		vm->reset();

		// restore inserted floppy disks
#ifdef USE_FD1
		for(int drv = 0; drv < MAX_DRIVE; drv++) {
			if(disk_insert[drv].path[0] != _T('\0')) {
				vm->open_disk(drv, disk_insert[drv].path, disk_insert[drv].offset);
			}
		}
#endif
		cpu_clock_low = config.cpu_clock_low;
	}
	else {
#endif
		// reset virtual machine
		vm->reset();
#ifdef USE_CPU_CLOCK_LOW
	}
#endif

#ifdef USE_MEDIA
	stop_media();
#endif
	// restart recording
//	restart_rec_video();
//	restart_rec_sound();
}

#ifdef USE_SPECIAL_RESET
void EMU::special_reset()
{
	// reset virtual machine
	vm->special_reset();
#ifdef USE_MEDIA
	stop_media();
#endif
	// restart recording
//	restart_rec_video();
//	restart_rec_sound();
}

void EMU::warm_reset()
{
	// reset virtual machine
	vm->warm_reset();
#ifdef USE_MEDIA
	stop_media();
#endif
}
#endif

void EMU::change_dipswitch(int num)
{
	config.dipswitch ^= (1 << num);
	vm->change_dipswitch(num);
}

#ifdef USE_POWER_OFF
void EMU::notify_power_off()
{
	vm->notify_power_off();
}
#endif

bool EMU::get_pause(int idx) {
	switch(idx) {
		case 1:
			return vm_syspause;
			break;
		case 2:
			return vm_usrpause;
			break;
	}
	return vm_pause;
}
void EMU::set_pause(int idx, bool val) {
	switch(idx) {
		case 1:
			vm_syspause = val;
			break;
		case 2:
			vm_usrpause = val;
			out_info(_T("Pause"), val, -1);

			break;
	}
	vm_pause = (vm_syspause | vm_usrpause);

	// mute sound
	mute_sound(vm_pause);
}
bool *EMU::get_pause_ptr(void) {
	return &vm_pause;
}

void EMU::application_path(_TCHAR *path)
{
	_tcscpy(path, app_path);
}
const _TCHAR *EMU::application_path()
{
	return app_path;
}
void EMU::initialize_path(_TCHAR *path)
{
	_tcscpy(path, ini_path);
}
const _TCHAR *EMU::initialize_path()
{
	return ini_path;
}

void EMU::set_gui(GUI *new_gui)
{
	gui = new_gui;

}

// ----------------------------------------------------------------------------

void EMU::send_log_message(int level, const _TCHAR *levelstr, const _TCHAR *msg)
{
#ifdef _BML3MK5
	if (msgboard && level != LOG_DEBUG) {
		msgboard->SetMessageF(_T("[%s] %s"), levelstr, msg);
	}
#endif
}

// ----------------------------------------------------------------------------

void EMU::out_msg(const _TCHAR* msg, bool set, int sec)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetMessage(msg, sec);
		} else {
			msgboard->DeleteMessage(msg);
		}
	}
#endif
}
void EMU::out_msg(const _TCHAR* msg, bool set)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetMessage(msg);
		} else {
			msgboard->DeleteMessage(msg);
		}
	}
#endif
}

void EMU::out_info(const _TCHAR* msg, bool set, int sec)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetInfo(msg, sec);
		} else {
			msgboard->DeleteInfo(msg);
		}
	}
#endif
}

void EMU::out_info(const _TCHAR* msg, bool set)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetInfo(msg);
		} else {
			msgboard->DeleteInfo(msg);
		}
	}
#endif
}

void EMU::out_infof(const _TCHAR* format, ...)
{
#ifdef _BML3MK5
	if (msgboard) {
		_TCHAR buffer[1024];

		va_list ap;

		va_start(ap, format);
		_vstprintf(buffer, format, ap);
		va_end(ap);

		msgboard->SetInfo(buffer);
	}
#endif
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------

void EMU::change_cpu_power(int num)
{
	config.cpu_power = num;

	if (num == 0) {
		out_infof(_T("CPU x0.5"));
	} else {
		out_infof(_T("CPU x%d"), (1 << (num - 1)));
	}
	update_config();
}
#ifdef _BML3MK5
void EMU::change_sync_irq()
{
	config.sync_irq = !config.sync_irq;

	out_infof(_T("Sync IRQ %s"),(config.sync_irq ? _T("ON") : _T("OFF")));
	update_config();
}
#endif
#ifdef USE_CART
void EMU::open_cart(_TCHAR* file_path)
{
	if (!vm) return;

	vm->open_cart(file_path);

	// restart recording
	restart_rec_video();
	restart_rec_sound();
}
void EMU::close_cart()
{
	if (!vm) return;

	vm->close_cart();

	// stop recording
	stop_rec_video();
	stop_rec_sound();
}
#endif

#ifdef USE_FD1
bool EMU::open_disk_main(int drv, _TCHAR* file_path, int offset, uint32 flags)
{
	if(disk_inserted(drv) || disk_insert[drv].wait_count != 0) {
		if(disk_inserted(drv)) {
			vm->close_disk(drv);
#ifdef _BML3MK5
		}
		return vm->open_disk(drv, file_path, offset, flags);
#else
			// wait 0.5sec
#ifdef SUPPORT_VARIABLE_TIMING
			disk_insert[drv].wait_count = (int)(vm->frame_rate() / 2);
#else
			disk_insert[drv].wait_count = (int)(FRAMES_PER_SEC / 2);
#endif
		}
		_tcscpy(disk_insert[drv].path, file_path);
		disk_insert[drv].offset = offset;
		disk_insert[drv].flags = flags;

		return true;
#endif
	}
	else {
		return vm->open_disk(drv, file_path, offset, flags);
	}
}
bool EMU::open_disk(int drv, _TCHAR* file_path, int bank_num, int offset, uint32 flags)
{
	bool rc = false;
	memset(&config.opened_disk_path[drv], 0, sizeof(pathex_t));
	if (vm && file_path && file_path[0] != 0) {
		if (open_disk_main(drv, file_path, offset, flags)) {
			pathex_t buf;
			memset(&buf, 0, sizeof(buf));
			_tcscpy(buf.path, file_path);
			buf.num = bank_num;
			UPDATE_HISTORY_EX(buf, config.recent_disk_path[drv], config.updated_recent_disk_path[drv]);
			memcpy(&config.opened_disk_path[drv], &buf, sizeof(pathex_t));
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Floppy image couldn't be opened."));
		}
		copy_initial_path(config.initial_disk_path, file_path);
	}
	return rc;
}
bool EMU::open_disk_by_bank_num(int drv, _TCHAR* file_path, int bank_num, uint32 flags, bool multiopen)
{
	bool rc = false;

	d88_file[drv].bank_nums = 0;
	d88_file[drv].cur_bank = -1;
	d88_file[drv].bank[0].offset = 0;
	d88_file[drv].prev_bank = -2;

	if(UTILITY::check_file_extension(file_path, _T(".d88")) || UTILITY::check_file_extension(file_path, _T(".d77"))) {
		FILEIO *fio = new FILEIO();
		if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
			try {
				fio->Fseek(0, FILEIO_SEEK_END);
				int file_size = (int)fio->Ftell();
				int file_offset = 0;
				while(file_offset + 0x2b0 <= file_size && d88_file[drv].bank_nums < MAX_D88_BANKS) {
					d88_file[drv].bank[d88_file[drv].bank_nums].offset = file_offset;
					fio->Fseek(file_offset, FILEIO_SEEK_SET);
#ifdef _UNICODE
					char tmp[18];
					fio->Fread(tmp, 17, 1);
					tmp[17] = 0;
					UTILITY::mbs_to_wcs(d88_file[drv].bank[d88_file[drv].bank_nums].name, tmp, 18);
#else
					fio->Fread(d88_file[drv].bank[d88_file[drv].bank_nums].name, 17, 1);
					d88_file[drv].bank[d88_file[drv].bank_nums].name[17] = 0;
#endif
					fio->Fseek(file_offset + 0x1c, FILEIO_SEEK_SET);
					file_offset += fio->Fgetc();
					file_offset += fio->Fgetc() << 8;
					file_offset += fio->Fgetc() << 16;
					file_offset += fio->Fgetc() << 24;
					d88_file[drv].bank_nums++;
				}
				_tcscpy(d88_file[drv].path, file_path);
				if (d88_file[drv].bank_nums > bank_num) {
					d88_file[drv].cur_bank = bank_num;
				} else {
					d88_file[drv].cur_bank = 0;
					bank_num = 0;
				}
			}
			catch(...) {
				d88_file[drv].bank_nums = 0;
				bank_num = 0;
			}
		}
		delete fio;
	}
	rc = open_disk(drv, file_path, bank_num, d88_file[drv].bank[bank_num].offset, flags);
	if (!rc || !multiopen) return rc;
#ifdef USE_FD2
	if(drv == 0 && bank_num + 1 < d88_file[drv].bank_nums) {
		rc = open_disk_by_bank_num(drv + 1, file_path, bank_num + 1, flags, multiopen);
		if (!rc) return rc;
	}
#endif
#ifdef USE_FD4
	if(drv == 2 && bank_num + 1 < d88_file[drv].bank_nums) {
		rc = open_disk_by_bank_num(drv + 1, file_path, bank_num + 1, flags, multiopen);
		if (!rc) return rc;
	}
#endif
#ifdef USE_FD6
	if(drv == 4 && bank_num + 1 < d88_file[drv].bank_nums) {
		rc = open_disk_by_bank_num(drv + 1, file_path, bank_num + 1, flags, multiopen);
		if (!rc) return rc;
	}
#endif
	return rc;
}
bool EMU::open_disk_with_sel_bank(int drv, int bank_num)
{
	bool rc = false;
	if(d88_file[drv].cur_bank != bank_num) {
		if ((rc = open_disk(drv, d88_file[drv].path, bank_num, d88_file[drv].bank[bank_num].offset, 0)) == true) {
			d88_file[drv].cur_bank = bank_num;
			d88_file[drv].prev_bank = -2;
		}
	}
	return rc;
}
void EMU::close_disk(int drv)
{
	if (!vm) return;

	memset(&config.opened_disk_path[drv], 0, sizeof(pathex_t));

	vm->close_disk(drv);

//	memset(&d88_file[drv], 0, sizeof(d88_file_t));
	d88_file[drv].cur_bank = -1;
	d88_file[drv].prev_bank = -2;
}
void EMU::initialize_disk_insert()
{
	memset(disk_insert, 0, sizeof(disk_insert));
}
void EMU::update_disk_insert()
{
	if (!vm) return;

	for(int drv = 0; drv < MAX_DRIVE; drv++) {
		if(disk_insert[drv].wait_count != 0 && --disk_insert[drv].wait_count == 0) {
			vm->open_disk(drv, disk_insert[drv].path, disk_insert[drv].offset, disk_insert[drv].flags);
		}
	}
}
int EMU::change_disk(int drv)
{
	return (vm ? vm->change_disk(drv) : 0);
}
int EMU::get_disk_side(int drv)
{
	return (vm ? vm->get_disk_side(drv) : 0);
}
void EMU::toggle_disk_write_protect(int drv)
{
	if (!vm) return;

	vm->toggle_disk_write_protect(drv);
}
bool EMU::disk_write_protected(int drv)
{
	return (vm ? vm->disk_write_protected(drv) : false);
}
bool EMU::disk_inserted(int drv)
{
	return (vm ? vm->disk_inserted(drv) : false);
}
bool EMU::changed_cur_bank(int drv)
{
	bool upd = (d88_file[drv].cur_bank != d88_file[drv].prev_bank);
	d88_file[drv].prev_bank = d88_file[drv].cur_bank;
	return upd;
}
#endif

#ifdef USE_QUICKDISK
void EMU::open_quickdisk(_TCHAR* file_path)
{
	if (!vm) return;

	vm->open_quickdisk(file_path);
}
void EMU::close_quickdisk()
{
	if (!vm) return;

	vm->close_quickdisk();
}
#endif

#ifdef USE_DATAREC
bool EMU::play_datarec(_TCHAR* file_path)
{
	bool rc = false;
	memset(&config.opened_datarec_path, 0, sizeof(pathex_t));
	if (vm && file_path && file_path[0] != 0) {
		if (vm->play_datarec(file_path)) {
			pathex_t buf;
			memset(&buf, 0, sizeof(buf));
			_tcscpy(buf.path, file_path);
			UPDATE_HISTORY_EX(buf, config.recent_datarec_path, config.updated_recent_datarec_path);
			memcpy(&config.opened_datarec_path, &buf, sizeof(pathex_t));
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Tape image couldn't be opened."));
		}
		copy_initial_path(config.initial_datarec_path, file_path);
	}

	return rc;
}
bool EMU::rec_datarec(_TCHAR* file_path)
{
	bool rc = false;
	memset(config.opened_datarec_path.path, 0, sizeof(pathex_t));
	if (vm && file_path && file_path[0] != 0) {
		if (vm->rec_datarec(file_path)) {
			pathex_t buf;
			memset(&buf, 0, sizeof(buf));
			_tcscpy(buf.path, file_path);
			UPDATE_HISTORY_EX(buf, config.recent_datarec_path, config.updated_recent_datarec_path);
			memcpy(&config.opened_datarec_path, &buf, sizeof(pathex_t));
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Tape image couldn't be saved."));
		}
		copy_initial_path(config.initial_datarec_path, file_path);
	}

	return rc;
}
void EMU::close_datarec()
{
	if (!vm) return;

	memset(&config.opened_datarec_path, 0, sizeof(pathex_t));

	vm->close_datarec();
}
void EMU::rewind_datarec()
{
	if (!vm) return;

	vm->rewind_datarec();
}
void EMU::fast_forward_datarec()
{
	if (!vm) return;

	vm->fast_forward_datarec();
}
void EMU::realmode_datarec()
{
	if (!vm) return;

	vm->realmode_datarec();
}
bool EMU::datarec_opened(bool play_mode)
{
	return (vm ? vm->datarec_opened(play_mode) : false);
}
#endif

#ifdef USE_DATAREC_BUTTON
void EMU::push_play()
{
	if (!vm) return;

	vm->push_play();
}
void EMU::push_stop()
{
	if (!vm) return;

	vm->push_stop();
}
#endif

#ifdef USE_BINARY_FILE1
void EMU::load_binary(int drv, _TCHAR* file_path)
{
	vm->load_binary(drv, file_path);
}
void EMU::save_binary(int drv, _TCHAR* file_path)
{
	vm->save_binary(drv, file_path);
}
#endif
#ifdef USE_SCANLINE
void EMU::change_screen_scanline()
{
	config.scan_line = !config.scan_line;

	out_infof(_T("Scanline %s"),(config.scan_line ? _T("ON") : _T("OFF")));
	update_config();
}
#endif
#ifdef _BML3MK5
void EMU::change_screen_use_direct3d(int num)
{
	if(enable_direct3d) {
		uint8 pre_num = config.use_direct3d;

		if (num >= 0) {
			config.use_direct3d = (pre_num == num) ? 0 : num;
		} else {
			config.use_direct3d = (pre_num + 1) % 3;
		}

		if (config.use_direct3d != 0) {
			set_d3dpresent_interval();
			reset_d3device(main_window_handle);
		}

		if (pD3Device != NULL && pre_num == 0 && config.use_direct3d != 0) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, false);
		} else if (pre_num != 0 && config.use_direct3d == 0) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, true);
		}

		switch (config.use_direct3d) {
			case 0:
				out_infof(_T("Direct3D OFF"));
				break;
			case 1:
				out_infof(_T("Direct3D ON (Sync)"));
				break;
			case 2:
				out_infof(_T("Direct3D ON (Async)"));
				break;
		}
		update_config();
	}
}
void EMU::change_screen_afterimage(int num)
{
	if (num >= 0) {
		config.afterimage = (config.afterimage == num) ? 0 : num;
	} else {
		config.afterimage = (config.afterimage + 1) % 3;
	}

	if (config.afterimage != 0) {
		out_infof(_T("Afterimage%d ON"), config.afterimage);
	} else {
		out_info(_T("Afterimage OFF"));
	}
	update_config();
}
void EMU::change_screen_keepimage(int num)
{
	if (num >= 0) {
		config.keepimage = (config.keepimage == num) ? 0 : num;
	} else {
		config.keepimage = (config.keepimage + 1) % 3;
	}

	if (config.keepimage != 0) {
		out_infof(_T("Keepimage%d ON"), config.keepimage);
	} else {
		out_info(_T("Keepimage OFF"));
	}
	update_config();
}
#endif

bool EMU::now_skip()
{
	return (vm ? vm->now_skip() : false);
}

void EMU::update_config()
{
	if (!vm) return;

	vm->update_config();
}

#ifdef USE_PRINTER
bool EMU::save_printer(int dev, _TCHAR* file_path)
{
	bool rc = false;
	if (vm && file_path && file_path[0] != 0) {
		if (vm->save_printer(dev, file_path)) {
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Print image couldn't be saved."));
		}
		copy_initial_path(config.initial_printer_path, file_path);
	}
	return rc;
}
void EMU::clear_printer(int dev)
{
	if (!vm) return;

	vm->clear_printer(dev);
}
int  EMU::get_printer_buffer_size(int dev)
{
	return (vm ? vm->get_printer_buffer_size(dev) : 0);
}
uint8* EMU::get_printer_buffer(int dev)
{
	return (vm ? vm->get_printer_buffer(dev) : NULL);
}
void EMU::enable_printer_direct(int dev)
{
	if (!vm) return;

	vm->enable_printer_direct(dev);
}
bool EMU::print_printer(int dev)
{
	return (vm ? vm->print_printer(dev) : false);
}
#endif

#ifdef _BML3MK5
void EMU::enable_comm_server(int dev)
{
	if (!vm) return;

	vm->enable_comm_server(dev);
}
void EMU::enable_comm_connect(int dev)
{
	if (!vm) return;

	vm->enable_comm_connect(dev);
}

int EMU::update_led()
{
	return (vm ? vm->update_led() : 0);
}

void EMU::show_led()
{
	if (ledbox) {
		config.show_dlg ^= 1;
		ledbox->Show(config.show_dlg & 9);
		out_infof(_T("%s LED"), (config.show_dlg & 1) ? _T("Show") : _T("Hide"));
	}
}

void EMU::inside_led()
{
	if (ledbox) {
		config.show_dlg ^= 8;
		ledbox->Show(config.show_dlg & 9);
		out_infof(_T("%s LED"), (config.show_dlg & 8) ? _T("Inside") : _T("Outside"));
	}
}

void EMU::move_led()
{
	if (ledbox) {
		ledbox->Move();
	}
}

void EMU::change_pos_led(int num)
{
	if (ledbox) {
		if (num == -1) {
			config.led_pos = (config.led_pos + 1) % 4;
		} else {
			config.led_pos = num;
		}
		ledbox->SetPos(config.led_pos);
		ledbox->MoveInPlace();
	}
}

bool EMU::is_shown_ledbox(int id)
{
	switch(id) {
		case 0:
			return (config.show_dlg & 1) ? true : false;
			break;
		case 1:
			return (config.show_dlg & 8) ? true : false;
			break;
	}
	return false;
}

void EMU::show_message_board()
{
	if (msgboard) {
		config.show_dlg ^= 2;
		msgboard->SetVisible(config.show_dlg & 2 ? true : false);
		out_infof(_T("%s Message"), (config.show_dlg & 2) ? _T("Show") : _T("Hide"));
	}
}

bool EMU::is_shown_message_board()
{
	return (config.show_dlg & 2) ? true : false;
}

void EMU::toggle_use_joypad()
{
	config.show_dlg ^= 4;
	if (config.show_dlg & 4) {
		use_joystick = true;
		reset_joystick();
		out_infof(_T("Enable Joypad"));
	} else {
		release_joystick();
		use_joystick = false;
		out_infof(_T("Disable Joypad"));
	}
}

bool EMU::is_enable_joypad()
{
	return (config.show_dlg & 4) ? true : false;
}

void EMU::toggle_enable_lightpen(void)
{
	config.show_dlg ^= 0x10;
	if (config.show_dlg & 0x10) {
		out_infof(_T("Enable Lightpen"));
	} else {
		out_infof(_T("Disable Lightpen"));
	}
}
bool EMU::is_enable_lightpen(void)
{
	return (config.show_dlg & 0x10) ? true : false;
}

void EMU::save_keybind()
{
	if (!vm) return;

	vm->save_keybind();
	if (use_joystick) {
		reset_joystick();
	}
}

void EMU::change_fdd_type(int num, bool reset)
{
	_TCHAR msg[128];
	if (num == -1) {
		num = get_parami(ParamFddType);
		num = (num + 1) % 3;
	}
	set_parami(ParamFddType, num);
	int io_port = get_parami(ParamIOPort);
	switch(num) {
	case 0:
		_tcscpy(msg, _T("Non FDD Mode"));
		io_port &= ~3;
		break;
	case 1:
		_tcscpy(msg, _T("3inch compact FDD Mode"));
		io_port &= ~3;
		io_port |= 2;
		break;
	case 2:
		_tcscpy(msg, _T("5.25inch mini FDD Mode"));
		io_port &= ~3;
		io_port |= 1;
		break;
	default:
		_tcscpy(msg, _T("Unsupported FDD Mode"));
		break;
	}
	if (config.fdd_type != num && !reset) {
		_tcscat(msg, _T(" (Need PowerOn)"));
	}
	set_parami(ParamIOPort, io_port);
	if (reset) {
		config.fdd_type = num;
		config.io_port = io_port;
		out_log(LOG_INFO, msg);
	} else {
		out_infof(msg);
	}
}

bool EMU::save_state(_TCHAR* file_path)
{
	bool rc = false;
	memset(config.saved_state_path, 0, sizeof(config.saved_state_path));
	if (vm && file_path && file_path[0] != 0) {
		if (vm->save_state(file_path)) {
			_tcsncpy(config.saved_state_path, file_path, _MAX_PATH);
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Status image couldn't be saved."));
		}
		copy_initial_path(config.initial_state_path, file_path);
	}
	return rc;
}

bool EMU::load_state(_TCHAR* file_path)
{
	bool rc = false;
	if (vm && file_path && file_path[0] != 0) {
		if (vm->load_state(file_path)) {
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Status image couldn't be loaded."));
		}
		copy_initial_path(config.initial_state_path, file_path);
	}
	return rc;
}

#ifdef USE_KEY_RECORD
bool EMU::play_reckey(_TCHAR* file_path)
{
	bool rc = false;
	if (vm && file_path && file_path[0] != 0) {
		if (vm->play_reckey(file_path)) {
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Auto key file couldn't be opened."));
		}
		copy_initial_path(config.initial_state_path, file_path);
	}
	return rc;
}

bool EMU::record_reckey(_TCHAR* file_path)
{
	bool rc = false;
	if (vm && file_path && file_path[0] != 0) {
		if (vm->record_reckey(file_path)) {
			rc = true;
		} else {
			out_log(LOG_ERROR, _T("Record key file couldn't be saved."));
		}
		copy_initial_path(config.initial_state_path, file_path);
	}
	return rc;
}

void EMU::stop_reckey(bool stop_play, bool stop_record)
{
	vm->stop_reckey(stop_play, stop_record);
}
#endif

void EMU::update_ui(int flags)
{
}
int EMU::get_parami(int id)
{
	if (id < ParamiUnknown) {
		return parami[id];
	}
	return 0;
}

void EMU::set_parami(int id, int val)
{
	if (id < ParamiUnknown) {
		parami[id] = val;
	}
}

void *EMU::get_paramv(int id)
{
	if (id < ParamvUnknown) {
		return paramv[id];
	}
	return NULL;
}

void EMU::set_paramv(int id, void *val)
{
	if (id < ParamvUnknown) {
		paramv[id] = val;
	}
}

void EMU::update_params()
{
	change_fdd_type(get_parami(ParamFddType), true);
}

#endif	/* _BML3MK5 */

/// static functions

/**
 *	copy dir path
 */
void EMU::copy_initial_path(_TCHAR *dst, _TCHAR *file_path)
{
	_TCHAR dir_path[_MAX_PATH];

	UTILITY::get_dir_and_basename(file_path, dir_path, NULL);

	memset(dst, 0, _MAX_PATH * sizeof(_TCHAR));
	_tcsncpy(dst, dir_path, _MAX_PATH);
}

