/** @file utility.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.04.01 -

	@brief [ utility ]
*/

#include "utility.h"
#include <ctype.h>
#ifdef _UNICODE
#include "ConvertUTF.h"
#endif

#ifdef _WIN32
// #define USE_WINDOWS_API
#include <Windows.h>
#include <Shlwapi.h>

#define PATH_SEP_CHAR1	_T('\\')
#define PATH_SEP_STR1	_T("\\")
#define PATH_GPARENT_STR1	_T("...\\")
#define PATH_PARENT_STR1	_T("..\\")
#define PATH_CURRENT_STR1	_T(".\\")

#define PATH_SEP_CHAR2	_T('/')
#define PATH_SEP_STR2	_T("/")
#define PATH_GPARENT_STR2	_T(".../")
#define PATH_PARENT_STR2	_T("../")
#define PATH_CURRENT_STR2	_T("./")

#else	/* !_WIN32 */

#define PATH_SEP_CHAR1	_T('/')
#define PATH_SEP_STR1	_T("/")
#define PATH_GPARENT_STR1	_T(".../")
#define PATH_PARENT_STR1	_T("../")
#define PATH_CURRENT_STR1	_T("./")

#define PATH_SEP_CHAR2	_T('\\')
#define PATH_SEP_STR2	_T("\\")
#define PATH_GPARENT_STR2	_T("...\\")
#define PATH_PARENT_STR2	_T("..\\")
#define PATH_CURRENT_STR2	_T(".\\")

#endif

namespace UTILITY
{

// ----------------------------------------------------------------------
// file path conversion

/**
 *	match extension
 */
bool check_file_extension(const _TCHAR* file_path, const _TCHAR* ext)
{
	int nam_len = (int)_tcslen(file_path);
	int ext_len = (int)_tcslen(ext);

#ifdef _WIN32
	if(nam_len >= ext_len && _tcsnicmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#else
#ifdef _UNICODE
	if(nam_len >= ext_len && wcsncasecmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#else
	if(nam_len >= ext_len && strncasecmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#endif
#endif
		return true;
	}
	return false;
}

/**
 *	convert path separator
 */
void convert_path_separator(_TCHAR *path)
{
	size_t len = _tcslen(path);

	for(size_t i=0; i<len; i++) {
		if (path[i] == PATH_SEP_CHAR2) {
			path[i]=PATH_SEP_CHAR1;
		}
	}
}

/**
 *	add path separator
 */
void add_path_separator(_TCHAR *path)
{
	size_t len = _tcslen(path);

	if (len > 0 && len < (_MAX_PATH - 1)) {
		if (path[len-1] != PATH_SEP_CHAR1) {
			path[len]=PATH_SEP_CHAR1;
			path[len+1]=_T('\0');
		}
	}
}

/**
 *	get dirname and basename (pattern 1)
 */
void get_dir_and_basename(const _TCHAR *path, _TCHAR *dir, _TCHAR *name)
{
	const _TCHAR *p = NULL;

	if (dir != NULL)  {
		memset(dir,  0, _MAX_PATH * sizeof(_TCHAR));
	}
	if (name != NULL) {
		memset(name, 0, _MAX_PATH * sizeof(_TCHAR));
	}
	p = _tcsrchr(path, PATH_SEP_CHAR1);
	if (p != NULL) {
		if (dir  != NULL) {
			_tcsncpy(dir,  path, p-path+1);
			dir[p-path+1] = _T('\0');
		}
		if (name != NULL) {
			_tcsncpy(name, p+1, _MAX_PATH-1);
		}
	} else {
		if (name != NULL) {
			_tcsncpy(name, path, _MAX_PATH-1);
		}
	}
	return;
}

/**
 *	get dirname and basename (pattern 2)
 */
void get_dir_and_basename(_TCHAR *path, _TCHAR *name)
{
	_TCHAR *p = NULL;

	if (name != NULL) memset(name, 0, sizeof(_TCHAR) * _MAX_PATH);

	p = _tcsrchr(path, PATH_SEP_CHAR1);
	if (p != NULL) {
		if (name != NULL) _tcsncpy(name, p+1, _MAX_PATH-1);
		*(p+1) = _T('\0');
	} else {
		if (name != NULL) _tcsncpy(name, path, _MAX_PATH-1);
		*path = _T('\0');
	}
	return;
}

/**
 *	triming center of long string
 */
_TCHAR *trim_center(const _TCHAR *str, int maxlen)
{
	static _TCHAR buff[_MAX_PATH];

	int srclen = (int)_tcslen(str);
	memset(buff, 0, sizeof(buff));
	if (srclen > maxlen + 4) {
		int len = maxlen / 2 - 2;
		_tcsncpy(buff, str, len);
		_tcscat(buff, _T("..."));
		_tcscat(buff, &str[srclen-len]);
	} else {
		_tcscpy(buff, str);
	}
	return buff;
}

/**
 *	make relative path if match base_path
 */
bool make_relative_path(const _TCHAR *base_path, _TCHAR *path)
{
	_TCHAR new_path[_MAX_PATH];

#ifdef USE_WINDOWS_API	// Use Windows API //
	if (!PathRelativePathTo(new_path, base_path, FILE_ATTRIBUTE_DIRECTORY, path, 0)) {
		return false;
	}
	_tcscpy(path, new_path);

#else	// ! Use Windows API //
	_TCHAR s_base_path[_MAX_PATH];
	_TCHAR s_path[_MAX_PATH];

	_tcscpy(s_base_path, slim_path(base_path));
	_tcscpy(s_path, slim_path(path));

	int len1 = (int)_tcslen(s_base_path);

	if (len1 == 0) {
		return true;
	}
	if (_tcsncmp(s_path, s_base_path, 2) != 0) {
		// if first 2 chars are different, stop convert it
		return true;
	}

	const _TCHAR *p1st = s_base_path;
	const _TCHAR *p1ed = NULL;
	_TCHAR *p2st = s_path;
	int parent = 0;

	while (p1st < s_base_path + len1 && (p1ed = _tcschr(p1st, PATH_SEP_CHAR1)) != NULL) {
		int len_part = (int)(p1ed - s_base_path + 1);
		if (_tcsncmp(s_path, s_base_path, len_part) == 0) {
			// same
			p2st = s_path + len_part;
		} else {
			// different
			parent++;
		}
		p1st = p1ed + 1;
	}

	// add relative path
	memset(new_path, 0, sizeof(new_path));
	for(; parent > 0; parent--) {
		_tcscat(new_path, PATH_PARENT_STR1);
	}
	_tcscat(new_path, p2st);

	_tcscpy(path, new_path);

#endif	// end of Use Windows API //
	return true;
}

/**
 *	make absolute path
 */
bool make_absolute_path(const _TCHAR *base_path, _TCHAR *path)
{
	int len1 = (int)_tcslen(base_path);

	if (len1 == 0) {
		return false;
	}

	_TCHAR new_path[_MAX_PATH];

#ifdef USE_WINDOWS_API	// Use Windows API //
	if (!PathCombine(new_path, base_path, path)) {
		return false;
	}
	_tcscpy(path, new_path);

#else	// ! Use Windows API //

#ifdef _WIN32
	// drive letter or root
	if ((_istalpha(path[0]) && path[1] == _T(':')) || path[0] == PATH_SEP_CHAR1) {
#else
	// root
	if (path[0] == PATH_SEP_CHAR1) {
#endif
		return true;
	}

	_tcscpy(new_path, slim_path(base_path));

	int parent = 0;
	_TCHAR *p2st = slim_path(path, &parent);

	// trim base path
	_TCHAR *p1st = NULL;
	for(; parent >= 0; parent--) {
		p1st = _tcsrchr(new_path, PATH_SEP_CHAR1);
		if (p1st == NULL) {
			break;
		}
		*p1st = _T('\0');
	}
	if (p1st == NULL) {
		// parent of root dir ???
		// so root
		p1st = new_path;
#ifdef _WIN32
		// if drive letter
		if (_istalpha(p1st[0]) && p1st[1] == _T(':')) {
			p1st += 2;
		}
#endif
	}
	*p1st = PATH_SEP_CHAR1;
	p1st++;
	*p1st = _T('\0');

	// add
	_tcscpy(p1st, p2st);
	_tcscpy(path, new_path);
#endif	// end of Use Windows API //
	return true;
}

/**
 *	slim path
 */
_TCHAR *slim_path(const _TCHAR *path, int *parent_num)
{
	static _TCHAR new_path[_MAX_PATH];

	const _TCHAR *p2st = path;
	const _TCHAR *p2ed = path;
	const _TCHAR *p2sep = NULL;
	int len2 = (int)_tcslen(path);
	int parent = 0;
	struct poslen_st {
		int pos;
		int len;
	} child_stack[64];
	int child = 0;

	memset(new_path, 0, sizeof(new_path));

	// if root
	if (_tcsncmp(p2st, PATH_SEP_STR1, 1) == 0) {
		// push path
		if (child < 64) {
			child_stack[child].pos = 0;
			child_stack[child].len = 1;
			child++;
		}
		p2st++;
		p2ed = p2st;
	}

	// parse parent path
	while (p2st < path + len2) {
		if (_tcsncmp(p2st, PATH_GPARENT_STR1, 4) == 0) {
			child-=2;
			if (child < 0) {
				parent-=child;
				child=0;
			}
			p2st += 4;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_PARENT_STR1, 3) == 0) {
			child--;
			if (child < 0) {
				parent-=child;
				child=0;
			}
			p2st += 3;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_CURRENT_STR1, 2) == 0) {
			p2st += 2;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_SEP_STR1, 1) == 0) {
			p2st += 1;
			p2ed = p2st;
		} else if ((p2sep = _tcschr(p2st, PATH_SEP_CHAR1)) != NULL) {
			if (p2st < p2sep) {
				// push path
				if (child < 64) {
					child_stack[child].pos = (int)(p2st - path);
					child_stack[child].len = (int)(p2sep - p2st + 1);
					child++;
				}
				p2st = p2sep + 1;
			} else {
				break;
			}
		} else {
			break;
		}
	}
	if (parent_num == NULL) {
		for(; parent > 0; parent--) {
			_tcsncat(new_path, PATH_PARENT_STR1, 3);
		}
	} else {
		*parent_num = parent;
	}
	for(int i=0; i<child; i++) {
		_tcsncat(new_path, path + child_stack[i].pos, child_stack[i].len);
	}
	_tcscat(new_path, p2st);

	return new_path;
}

// ----------------------------------------------------------------------
// string conversion

#ifdef _UNICODE
	/* _UNICODE */

bool conv_tchar_to_utf8(const wchar_t *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
	UTF8  **dst_st = (UTF8  **)&dstStart;
	UTF8   *dst_ed = (UTF8  *)(dstStart + dstMaxSize);
	ConversionResult rc = conversionOK;
	if (sizeof(wchar_t) == 4) {
		UTF32 **src_st = (UTF32 **)&srcStart;
		UTF32  *src_ed = (UTF32 *)(srcStart + srcMaxSize);
		rc = ConvertUTF32toUTF8((const UTF32 **)src_st, (const UTF32 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else if (sizeof(wchar_t) == 2) {
		UTF16 **src_st = (UTF16 **)&srcStart;
		UTF16  *src_ed = (UTF16 *)(srcStart + srcMaxSize);
		rc = ConvertUTF16toUTF8((const UTF16 **)src_st, (const UTF16 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else {
		return false;
	}
	if (rc != conversionOK) {
		return false;
	}
	return true;
}

wchar_t *wconv_from_native_path(const wchar_t *str, int maxSize)
{
	static wchar_t buff[_MAX_PATH];
	memset(buff, 0, sizeof(buff));
	_tcsncpy(buff, str, maxSize > _MAX_PATH ? _MAX_PATH : maxSize);
	return buff;
}

char *cconv_from_native_path(const wchar_t *str, int maxSize)
{
	static char buff[_MAX_PATH];
	memset(buff, 0, sizeof(buff));

#ifdef _WIN32
	int rc = 0;

	// WideChar -> UTF-8
	rc = WideCharToMultiByte(CP_UTF8, 0, str, -1, buff, _MAX_PATH, NULL, NULL);
#else
	bool rc = false;
	int len = (int)wcslen(str);
	len = len > maxSize ? maxSize : len;

	// WideChar -> UTF-8
	rc = conv_tchar_to_utf8(str, len, buff, _MAX_PATH);
#endif
	return buff;
}

int wcs_to_mbs(char *dst, const wchar_t *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);
#ifdef _WIN32
	int len = 0;

	// WideChar -> UTF-8
	len = WideCharToMultiByte(CP_UTF8, 0, src, -1, dst, maxSize, NULL, NULL);
#else
	int len = (int)wcslen(src);
	len = len > maxSize ? maxSize : len;

	// WideChar -> UTF-8
	conv_tchar_to_utf8(src, len, dst, maxSize);
#endif
	len = (int)strlen(dst);
	return len;
}

bool conv_utf8_to_tchar(const char *srcStart, int srcMaxSize, wchar_t *dstStart, int dstMaxSize)
{
	UTF8  **src_st = (UTF8  **)&srcStart;
	UTF8   *src_ed = (UTF8  *)(srcStart + srcMaxSize);
	ConversionResult rc = conversionOK;
	if (sizeof(wchar_t) == 4) {
		UTF32 **dst_st = (UTF32 **)&dstStart;
		UTF32  *dst_ed = (UTF32 *)(dstStart + dstMaxSize);
		rc = ConvertUTF8toUTF32((const UTF8 **)src_st, (const UTF8 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else if (sizeof(wchar_t) == 2) {
		UTF16 **dst_st = (UTF16 **)&dstStart;
		UTF16  *dst_ed = (UTF16 *)(dstStart + dstMaxSize);
		rc = ConvertUTF8toUTF16((const UTF8 **)src_st, (const UTF8 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else {
		return false;
	}
	if (rc != conversionOK) {
		return false;
	}
	return true;
}

wchar_t *conv_to_native_path(const wchar_t *str, int maxSize)
{
	static wchar_t buff[_MAX_PATH];
	memset(buff, 0, sizeof(buff));
	wcsncpy(buff, str, maxSize > _MAX_PATH ? _MAX_PATH : maxSize);
	// chomp lf
	size_t lastpos = wcslen(buff)-1;
	if (buff[lastpos-1] == L'\r' && buff[lastpos] == L'\n') {
		buff[lastpos-1] = L'\0';
	} else if (buff[lastpos] == L'\n' || buff[lastpos] == L'\r') {
		buff[lastpos] = L'\0';
	}
	return buff;
}

wchar_t *conv_to_native_path(const char *str, int maxSize)
{
	static wchar_t buff[_MAX_PATH];

	memset(buff, 0, sizeof(buff));
#ifdef _WIN32
	int rc = 0;

	// UTF-8 -> WideChar
	rc = MultiByteToWideChar(CP_UTF8, 0, str, -1, buff, _MAX_PATH);
#else
	bool rc = false;
	int len = (int)strlen(str);
	len = len > maxSize ? maxSize : len;

	// UTF-8 -> WideChar
	rc = conv_utf8_to_tchar(str, len, buff, _MAX_PATH);
#endif
	// chomp lf
	size_t lastpos = wcslen(buff)-1;
	if (buff[lastpos-1] == L'\r' && buff[lastpos] == L'\n') {
		buff[lastpos-1] = L'\0';
	} else if (buff[lastpos] == '\n' || buff[lastpos] == L'\r') {
		buff[lastpos] = L'\0';
	}
	return buff;
}

int mbs_to_wcs(wchar_t *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(wchar_t) * maxSize);
#ifdef _WIN32
	int len = 0;

	// MBS -> WideChar
	len = MultiByteToWideChar(CP_ACP, 0, src, -1, dst, maxSize);
#else
	int len = strlen(src);
	len = len > maxSize ? maxSize : len;

	// UTF-8 -> WideChar
	conv_utf8_to_tchar(src, len, dst, maxSize);
#endif
	// chomp lf
	size_t lastpos = wcslen(dst)-1;
	if (dst[lastpos-1] == L'\r' && dst[lastpos] == L'\n') {
		dst[lastpos-1] = L'\0';
	} else if (dst[lastpos] == L'\n' || dst[lastpos] == L'\r') {
		dst[lastpos] = L'\0';
	}
	len = (int)wcslen(dst);
	return len;
}

#else	/* !_UNICODE */

bool conv_tchar_to_utf8(const char *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[_MAX_PATH];
	// MBS -> WideChar
	rc = MultiByteToWideChar(CP_ACP, 0, srcStart, -1, dstTemp, _MAX_PATH);
	if (rc > 0) {
		// WideChar -> UTF-8
		rc = WideCharToMultiByte(CP_UTF8, 0, dstTemp, -1, dstStart, dstMaxSize, NULL, NULL);
	}
	if (rc == 0) {
		// cannot convert
		strncpy(dstStart, srcStart, dstMaxSize);
		return false;
	}
#else
	strncpy(dstStart, srcStart, dstMaxSize);
#endif
	return true;
}

char *cconv_from_native_path(const char *str, int maxSize)
{
	static char buff[_MAX_PATH];
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[_MAX_PATH];

	memset(buff, 0, sizeof(buff));
	// MBS -> WideChar
	rc = MultiByteToWideChar(CP_ACP, 0, str, -1, dstTemp, _MAX_PATH);
	if (rc > 0) {
		// WideChar -> UTF-8
		rc = WideCharToMultiByte(CP_UTF8, 0, dstTemp, -1, buff, _MAX_PATH, NULL, NULL);
	}
#else
	strncpy(buff, str, _MAX_PATH);
#endif
	return buff;
}

int wcs_to_mbs(char *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);

	strncpy(dst, src, maxSize);

	int len = (int)strlen(dst);
	return len;
}

bool conv_utf8_to_tchar(const char *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[_MAX_PATH];
	// UTF-8 -> WideChar
	rc = MultiByteToWideChar(CP_UTF8, 0, srcStart, -1, dstTemp, _MAX_PATH);
	if (rc > 0) {
		// WideChar -> MBS
		rc = WideCharToMultiByte(CP_ACP, 0, dstTemp, -1, dstStart, dstMaxSize, NULL, NULL);
	}
	if (rc == 0) {
		// cannot convert
		strncpy(dstStart, srcStart, dstMaxSize);
		return false;
	}
#else
	strncpy(dstStart, srcStart, dstMaxSize);
#endif
	return true;
}

char *conv_to_native_path(const char *str, int maxSize)
{
	static char buff[_MAX_PATH];
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[_MAX_PATH];

	memset(buff, 0, sizeof(buff));
	// UTF-8 -> WideChar
	rc = MultiByteToWideChar(CP_UTF8, 0, str, -1, dstTemp, _MAX_PATH);
	if (rc > 0) {
		// WideChar -> MBS
		rc = WideCharToMultiByte(CP_ACP, 0, dstTemp, -1, buff, _MAX_PATH, NULL, NULL);
	}
#else
	strncpy(buff, str, _MAX_PATH);
#endif
	// chomp lf
	size_t lastpos = strlen(buff)-1;
	if (buff[lastpos-1] == '\r' && buff[lastpos] == '\n') {
		buff[lastpos-1] = '\0';
	} else if (buff[lastpos] == '\n' || buff[lastpos] == '\r') {
		buff[lastpos] = '\0';
	}
	return buff;
}

int mbs_to_wcs(char *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);

	strncpy(dst, src, maxSize);

	// chomp lf
	size_t lastpos = strlen(dst)-1;
	if (dst[lastpos-1] == '\r' && dst[lastpos] == '\n') {
		dst[lastpos-1] = '\0';
	} else if (dst[lastpos] == '\n' || dst[lastpos] == '\r') {
		dst[lastpos] = '\0';
	}
	int len = (int)strlen(dst);
	return len;
}

#endif	/* _UNICODE */


}	// end of UTILITY
