/** @file event.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.11.29-

	@brief [ event manager ]
*/

#include "event.h"
#include "../config.h"

#ifndef EVENT_CONTINUOUS_SOUND
//#ifdef PCM1BIT_HIGH_QUALITY
#define EVENT_CONTINUOUS_SOUND
//#endif
#endif

void EVENT::initialize()
{
	// load config
	if(!(0 <= config.cpu_power && config.cpu_power <= 5)) {
		config.cpu_power = 1;
	}

#ifdef _BML3MK5
	if (config.sync_irq) {
		cpu_power = 1;
		event_power = config.cpu_power;
	} else {
		cpu_power = config.cpu_power;
		event_power = 1;
	}
	vm_pause = emu->get_pause_ptr();

#else
	cpu_power = config.cpu_power;
	event_power = 0;
#endif

	memset(vclocks, 0, sizeof(vclocks));
	memset(vline_split, 0, sizeof(vline_split));

	frame_split_num = 0;

	// initialize sound buffer
	sound_buffer = NULL;
	sound_tmp = NULL;

	now_rec_sound = emu->get_now_rec_sound_ptr();
}

void EVENT::initialize_sound(int rate, int samples)
{
	// initialize sound
	sound_rate = rate;
	sound_samples = samples;
#ifdef EVENT_CONTINUOUS_SOUND
	sound_tmp_samples = samples * 3;
#else
	sound_tmp_samples = samples;
#endif
	update_samples = (int)(1024.0 * (double)sound_rate / frames_per_sec / (double)lines_per_frame + 0.5);
//	update_config();
#ifdef USE_AUDIO_U8
	sound_buffer = (uint8*)malloc(sound_samples * sizeof(uint8) * 2);
	memset(sound_buffer, 0, sound_samples * sizeof(uint8) * 2);
	sound_empty = (uint8*)malloc(sound_samples * sizeof(uint8) * 2);
	memset(sound_empty, 128, sound_samples * sizeof(uint8) * 2);
#else
	sound_buffer = (uint16*)malloc(sound_samples * sizeof(uint16) * 2);
	memset(sound_buffer, 0, sound_samples * sizeof(uint16) * 2);
	sound_empty = (uint16*)malloc(sound_samples * sizeof(uint16) * 2);
	memset(sound_empty, 0, sound_samples * sizeof(uint16) * 2);
#endif
	sound_tmp = (int32*)malloc(sound_tmp_samples * sizeof(int32) * 2);
	memset(sound_tmp, 0, sound_tmp_samples * sizeof(int32) * 2);
	sound_buffer_ptr = accum_samples = 0;
	sound_buffer_overflow = 0;
	master_volume = 0;
#ifdef _DEBUG_SOUND_LATE
	sound_tmp_clocks = (uint64*)malloc(sound_tmp_samples * sizeof(uint64));
	memset(sound_tmp_clocks, 0, sound_tmp_samples * sizeof(uint64));
#endif
}

void EVENT::release()
{
	// release sound
	if(sound_buffer) {
		free(sound_buffer);
		sound_buffer = NULL;
	}
	if(sound_empty) {
		free(sound_empty);
		sound_empty = NULL;
	}
	if(sound_tmp) {
		free(sound_tmp);
		sound_tmp = NULL;
	}
#ifdef _DEBUG_SOUND_LATE
	if (sound_tmp_clocks) {
		free(sound_tmp_clocks);
		sound_tmp_clocks = NULL;
	}
#endif
}

void EVENT::reset()
{
	// clear events except loop event
	for(int i = 0; i < MAX_EVENT; i++) {
		if(event[i].active && event[i].loop_clock == 0) {
			cancel_event(i);
		}
	}

	event_remain = 0;
	cpu_remain = cpu_accum = cpu_done = 0;

	// reset sound
	if(sound_buffer) {
#ifdef USE_AUDIO_U8
		memset(sound_buffer, 0, sound_samples * sizeof(uint8) * 2);
#else
		memset(sound_buffer, 0, sound_samples * sizeof(uint16) * 2);
#endif
	}
	if(sound_tmp) {
		memset(sound_tmp, 0, sound_tmp_samples * sizeof(int32) * 2);
	}
//	sound_buffer_ptr = 0;

#ifdef _DEBUG_LOG
	initialize_done = true;
#endif
}

void EVENT::drive(int split_num)
{
	int v;
	int draw_count = ((1 << event_power) / FRAME_SPLIT_NUM / 2) * split_num;

	for(int epow = (1 << event_power); epow > 0; epow--) {
		if (frame_split_num == 0) {
			// raise pre frame events to update timing settings
			for(int i = 0; i < frame_event_count; i++) {
				frame_event[i]->event_pre_frame();
			}
			// generate clocks per line
			if(frames_per_sec != next_frames_per_sec || lines_per_frame != next_lines_per_frame) {
				frames_per_sec = next_frames_per_sec;
				lines_per_frame = next_lines_per_frame;

				int sum = (int)((double)d_cpu[0].cpu_clocks / frames_per_sec + 0.5);
				int remain = sum;

				for(int i = 0; i < lines_per_frame; i++) {
					vclocks[i] = (sum / lines_per_frame);
					remain -= vclocks[i];
				}
				for(int i = 0; i < remain; i++) {
					int index = (lines_per_frame * i / remain);
					vclocks[index]++;
				}
				sum = lines_per_frame;
				remain = sum;
				for(int sp = 1; sp <= FRAME_SPLIT_NUM*2; sp++) {
					vline_split[sp] = (sum / FRAME_SPLIT_NUM / 2);
					remain -= vline_split[sp];
				}
				for(int sp = 0; sp < remain; sp++) {
					int index = FRAME_SPLIT_NUM * 2 * sp / remain;
					vline_split[index+1]++;
				}
				for(int sp = 1; sp <= FRAME_SPLIT_NUM*2; sp++) {
					vline_split[sp] += vline_split[sp-1];
				}
				for(int i = 1; i < dcount_cpu; i++) {
					d_cpu[i].update_clocks = (int)(1024.0 * (double)d_cpu[i].cpu_clocks / (double)d_cpu[0].cpu_clocks + 0.5);
				}
				for(DEVICE* device = vm->first_device; device; device = device->next_device) {
					if(device->event_manager_id() == this_device_id) {
						device->update_timing(d_cpu[0].cpu_clocks, frames_per_sec, lines_per_frame);
					}
				}
				update_samples = (int)(1024.0 * (double)sound_rate / frames_per_sec / (double)lines_per_frame + 0.5);
			}

			// run virtual machine for 1 frame period
			for(int i = 0; i < frame_event_count; i++) {
				frame_event[i]->event_frame();
			}
		}
		if (!(*vm_pause)) {
			for(v = vline_split[frame_split_num]; v < vline_split[frame_split_num+1]; v++) {
				// run virtual machine per line
				for(int i = 0; i < vline_event_count; i++) {
					vline_event[i]->event_vline(v, vclocks[v]);
				}
				if(event_remain < 0) {
					if(-event_remain > vclocks[v]) {
						update_event(vclocks[v]);
					}
					else {
						update_event(-event_remain);
					}
				}
				event_remain += vclocks[v];
				cpu_remain += (vclocks[v] << cpu_power) >> 1;

				while(event_remain > 0) {
					int event_done = event_remain;
					if(cpu_remain > 0) {
						// run one opecode on primary cpu
						int cpu_done_tmp;
						if(dcount_cpu == 1) {
							cpu_done_tmp = d_cpu[0].device->run(-1);
						}
						else {
							// sync to sub cpus
							if(cpu_done == 0) {
								cpu_done = d_cpu[0].device->run(-1);
							}
							cpu_done_tmp = (cpu_done < 4) ? cpu_done : 4;
							cpu_done -= cpu_done_tmp;

							for(int i = 1; i < dcount_cpu; i++) {
								// run sub cpus
								d_cpu[i].accum_clocks += d_cpu[i].update_clocks * cpu_done_tmp;
								int sub_clock = d_cpu[i].accum_clocks >> 10;
								if(sub_clock) {
									d_cpu[i].accum_clocks -= sub_clock << 10;
									d_cpu[i].device->run(sub_clock);
								}
							}
						}
//#ifdef _BML3MK5
//						for(int i = 0; i < dcount_display; i++) {
//							d_display[i]->update_vram(v, cpu_done_tmp);
//						}
//#endif
						cpu_remain -= cpu_done_tmp;
						cpu_accum += cpu_done_tmp;
						event_done = (cpu_accum << 1) >> cpu_power;
						cpu_accum -= ((event_done << cpu_power) >> 1);
					}
					if(event_done > 0) {
						if(event_done > event_remain) {
							update_event(event_remain);
						}
						else {
							update_event(event_done);
						}
						event_remain -= event_done;
					}
				}
#ifdef _BML3MK5
				for(int i = 0; i < dcount_display; i++) {
					d_display[i]->update_vram(v, vclocks[v]);
				}
#endif
				update_sound(v);
			}
			if (v == lines_per_frame) {
				// draw screen if need
				if (draw_count < 2) {
					emu->draw_screen();
				}
				draw_count++;
				// update autokey if need
				emu->update_autokey();
			}
		} else {
			// now pausing
			for(v = vline_split[frame_split_num]; v < vline_split[frame_split_num+1]; v++) {
				update_sound(v);
			}
		}
		frame_split_num = (frame_split_num + 1) & ((FRAME_SPLIT_NUM << 1) - 1);
	}
}

void EVENT::update_event(int clock)
{
	uint64 event_clocks_tmp = event_clocks + clock;

	while(first_fire_event != NULL && first_fire_event->expired_clock <= event_clocks_tmp) {
		event_t *event_handle = first_fire_event;

		first_fire_event = event_handle->next;
		if(first_fire_event != NULL) {
			first_fire_event->prev = NULL;
		}
		if(event_handle->loop_clock != 0) {
			event_handle->expired_clock += event_handle->loop_clock;
			insert_event(event_handle);
		}
		else {
			event_handle->active = false;
			event_handle->next = first_free_event;
			first_free_event = event_handle;
		}
		event_clocks = event_handle->expired_clock;
		event_handle->device->event_callback(event_handle->event_id, 0);
	}
	event_clocks = event_clocks_tmp;
}

uint64 EVENT::current_clock()
{
	return event_clocks;
}

uint64 EVENT::passed_clock(uint64 prev)
{
	uint64 current = current_clock();
	return (current > prev) ? current - prev : current + (0xffffffffffffffffLL - prev) + 1;
}

uint32 EVENT::get_cpu_pc(int index)
{
	return d_cpu[index].device->get_pc();
}

void EVENT::register_event(DEVICE* device, int event_id, double usec, bool loop, int* register_id)
{
	int clock = (int)((double)d_cpu[0].cpu_clocks / 1000000.0 * usec + 0.5);
	register_event_by_clock(device, event_id, clock, loop, register_id);
}

void EVENT::register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id)
{
#ifdef _DEBUG_LOG
	if(!initialize_done && !loop) {
		emu->out_debug(_T("EVENT: non-loop event is registered before initialize is done"));
	}
#endif

	// register event
	if(first_free_event == NULL) {

		emu->out_log(LOG_ERROR,_T("EVENT: too many events !!! device_id:%d event_id:%d")
			,device->this_device_id,event_id);

		if(register_id != NULL) {
			*register_id = -1;
		}
		return;
	}
	event_t *event_handle = first_free_event;
	first_free_event = first_free_event->next;

	if(register_id != NULL) {
		*register_id = event_handle->index;
	}
	event_handle->active = true;
	event_handle->device = device;
	event_handle->event_id = event_id;
	event_handle->expired_clock = event_clocks + clock;
	event_handle->loop_clock = loop ? clock : 0;

	insert_event(event_handle);
}

void EVENT::insert_event(event_t *event_handle)
{
//	emu->out_debug(_T("EVENT: insert_event: device_id:%d event_id:%d register_id:%d")
//		,event_handle->device->this_device_id,event_handle->event_id,event_handle->index);
	if(first_fire_event == NULL) {
		first_fire_event = event_handle;
		event_handle->prev = event_handle->next = NULL;
	}
	else {
		for(event_t *insert_pos = first_fire_event; insert_pos != NULL; insert_pos = insert_pos->next) {
			if(insert_pos->expired_clock > event_handle->expired_clock) {
				if(insert_pos->prev != NULL) {
					// insert
					insert_pos->prev->next = event_handle;
					event_handle->prev = insert_pos->prev;
					event_handle->next = insert_pos;
					insert_pos->prev = event_handle;
					break;
				}
				else {
					// add to head
					first_fire_event = event_handle;
					event_handle->prev = NULL;
					event_handle->next = insert_pos;
					insert_pos->prev = event_handle;
					break;
				}
			}
			else if(insert_pos->next == NULL) {
				// add to tail
				insert_pos->next = event_handle;
				event_handle->prev = insert_pos;
				event_handle->next = NULL;
				break;
			}
		}
	}
}

void EVENT::cancel_event(int register_id)
{
	// cancel registered event
	if(0 <= register_id && register_id < MAX_EVENT) {
		event_t *event_handle = &event[register_id];
		if(event_handle->active) {
//			emu->out_debug(_T("EVENT: cancel_event: device_id:%d event_id:%d register_id:%d")
//				,event_handle->device->this_device_id,event_handle->event_id,event_handle->index);
			if(event_handle->prev != NULL) {
				event_handle->prev->next = event_handle->next;
			}
			else {
				first_fire_event = event_handle->next;
			}
			if(event_handle->next != NULL) {
				event_handle->next->prev = event_handle->prev;
			}
			event_handle->device = NULL;
			event_handle->active = false;
			event_handle->next = first_free_event;
			first_free_event = event_handle;
		}
	}
}

void EVENT::register_frame_event(DEVICE* dev)
{
	if(frame_event_count < MAX_EVENT) {
		frame_event[frame_event_count++] = dev;
	}
	else {
		emu->out_log(LOG_ERROR,_T("EVENT: too many frame events !!!"));
	}
}

void EVENT::register_vline_event(DEVICE* dev)
{
	if(vline_event_count < MAX_EVENT) {
		vline_event[vline_event_count++] = dev;
	}
	else {
		emu->out_log(LOG_ERROR,_T("EVENT: too many vline events !!!"));
	}
}

void EVENT::mix_sound(int samples)
{
	int32 *p;

	if (!(*vm_pause)) {
		if(samples > 0) {
			emu->lock_sound_buffer();
			memset(sound_tmp + sound_buffer_ptr * 2, 0, samples * sizeof(int32) * 2);
			for(int i = 0; i < dcount_sound; i++) {
				d_sound[i]->mix(sound_tmp + sound_buffer_ptr * 2, samples);
			}
			// control volume
			p = sound_tmp + sound_buffer_ptr * 2;
			for(int i = 0; i < (samples << 1); i++) {
				*p = *p * master_volume / 100;
				// limit
				if ((*p) >= 32768) *p = 32767;
				else if ((*p) < -32768) *p = -32768;
				p++;
			}
			// record sound if need
			record_sound(samples);

			sound_buffer_ptr += samples;
			emu->unlock_sound_buffer();
		}
		else {
			// notify to sound devices
			emu->lock_sound_buffer();
			for(int i = 0; i < dcount_sound; i++) {
				d_sound[i]->mix(sound_tmp + sound_buffer_ptr * 2, 0);
			}
			emu->unlock_sound_buffer();
		}
	} else {
		// now pausing
		if(samples > 0) {
			emu->lock_sound_buffer();
			memset(sound_tmp + sound_buffer_ptr * 2, 0, samples * sizeof(int32) * 2);

			// record sound if need
			record_sound(samples);

			sound_buffer_ptr += samples;
			emu->unlock_sound_buffer();
		}
	}
}

void EVENT::update_sound(int v)
{
	accum_samples += update_samples;
	int samples = (accum_samples >> 9 >> event_power);
	accum_samples -= (samples << 9 << event_power);
//	int samples = (int)(update_sound_rate_f / lines_per_frame / frames_per_sec / 16.0);
//	update_sound_rate_f = sound_rate_f + (update_sound_rate_f - (int)(samples * lines_per_frame * frames_per_sec * 16.0));

	// mix sound
	if(sound_tmp_samples - sound_buffer_ptr < samples) {
		samples = sound_tmp_samples - sound_buffer_ptr;
		sound_buffer_overflow++;
	}
#ifdef _DEBUG_SOUND_LATE
//	emu->out_debug(_T("EVENT::update_sound: clock:%lld"), event_clocks);
	for(int i=0; i<samples; i++) {
		sound_tmp_clocks[sound_buffer_ptr+i] = event_clocks;
	}
#endif
//	emu->out_debug("EVENT::update_sound accum_samples:%d samples:%d sound_tmp_samples:%d sound_buffer_ptr:%d",accum_samples,samples,sound_tmp_samples,sound_buffer_ptr);
	mix_sound(samples);
}

/// @attention called by another thread
/// @param[in] extra_frames 0
/// @param[in] samples      requested samples by DirectSound
#ifdef USE_AUDIO_U8
uint8* EVENT::create_sound(int* extra_frames, int samples)
#else
uint16* EVENT::create_sound(int* extra_frames, int samples)
#endif
{
	int frames = 0;

#ifdef EVENT_CONTINUOUS_SOUND
//	// drive extra frames to fill the sound buffer
//	emu->out_debug("EVENT::create_sound %ld",timeGetTime());
//	while(sound_samples > sound_buffer_ptr && frames < 0) {
//		emu->out_debug(" frames:%d sound_samples:%d sound_buffer_ptr:%d",frames,sound_samples,sound_buffer_ptr);
//		drive();
//		frames++;
//	}
#else
	// fill sound buffer
	int samples = sound_samples - sound_buffer_ptr;
	mix_sound(samples);
#endif
#ifdef LOW_PASS_FILTER
	// low-pass filter
	for(int i = 0; i < samples - 1; i++) {
		sound_tmp[i * 2    ] = (sound_tmp[i * 2    ] + sound_tmp[i * 2 + 2]) / 2; // L
		sound_tmp[i * 2 + 1] = (sound_tmp[i * 2 + 1] + sound_tmp[i * 2 + 3]) / 2; // R
	}
#endif
#ifdef _DEBUG_SOUND_LATE
	emu->out_log(LOG_DEBUG,_T("EVENT::create_sound 1 samples:%d buf_ptr:%d ovflow:%d"),samples,sound_buffer_ptr,sound_buffer_overflow);
#endif
#ifdef EVENT_CONTINUOUS_SOUND
	//  skip if buffer is not fill
	if (samples > sound_buffer_ptr) {
		*extra_frames = frames;
		return sound_empty;
	}
#endif
#ifdef _DEBUG_SOUND_LATE
	emu->out_log(LOG_DEBUG,_T("EVENT::create_sound: bufclk:%lld-%lld nowclk:%lld d:%lld"),sound_tmp_clocks[0],sound_tmp_clocks[sound_buffer_ptr-1],event_clocks,event_clocks-sound_tmp_clocks[0]);
#endif
	// copy to buffer
	for(int i = 0; i < samples * 2; i++) {
#ifdef USE_AUDIO_U8
		sound_buffer[i] = (uint8)(sound_tmp[i] / 256) + 128;
#else
		sound_buffer[i] = (uint16)sound_tmp[i];
#endif
	}
	if(sound_buffer_ptr > samples && sound_buffer_overflow < 60) {
		if (sound_buffer_overflow < 10) {
			sound_buffer_ptr -= samples;
		} else {
			sound_buffer_ptr -= samples + samples / 2;
		}
		memcpy(sound_tmp, sound_tmp + samples * 2, sound_buffer_ptr * sizeof(int32) * 2);
#ifdef _DEBUG_SOUND_LATE
		memcpy(sound_tmp_clocks, sound_tmp_clocks + samples, sound_buffer_ptr * sizeof(uint64));
#endif
	}
	else {
		sound_buffer_ptr = 0;
	}
//	emu->out_debug("EVENT::create_sound 2 samples:%d buf_ptr:%d",sound_samples,sound_buffer_ptr);

	sound_buffer_overflow = 0;

	*extra_frames = frames;
	return sound_buffer;
}

void EVENT::set_volume(int volume)
{
	master_volume = volume;
}

void EVENT::record_sound(int samples)
{
	if (*now_rec_sound) {
		emu->record_rec_sound(&sound_tmp[sound_buffer_ptr << 1], samples);
	}
}

void EVENT::update_config()
{
#ifdef _BML3MK5
	if (config.sync_irq) {
		event_power = config.cpu_power;
		cpu_power = 1;
		cpu_accum = 0;
	} else {
		cpu_power = config.cpu_power;
		event_power = 1;
		cpu_accum = 0;
	}
#else
	if(cpu_power != config.cpu_power) {
		cpu_power = config.cpu_power;
		cpu_accum = 0;
	}
#endif
//	sound_rate_f = ((sound_rate << 5) >> event_power);
//	update_sound_rate_f = sound_rate_f;
}

// ----------------------------------------------------------------------------

void EVENT::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 1;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy events
	memset(&vm_state, 0, sizeof(vm_state));

	vm_state.event_clocks = event_clocks;

	for(int i=0; i<MAX_EVENT; i++) {
		if (event[i].device != NULL) {
			strncpy(vm_state.event[i].class_name, event[i].device->this_class_name, 12);
			strncpy(vm_state.event[i].identifier, event[i].device->this_identifier, 4);
		}
		vm_state.event[i].event_id = event[i].event_id;
		vm_state.event[i].expired_clock = event[i].expired_clock;
		vm_state.event[i].loop_clock = event[i].loop_clock;
		vm_state.event[i].active = event[i].active ? 1 : 0;
		vm_state.event[i].index = event[i].index;

		vm_state.event[i].next_index = -1;
		for(int j=0; j<MAX_EVENT; j++) {
			if (event[i].next == &event[j]) {
				vm_state.event[i].next_index = j;
				break;
			}
		}
		vm_state.event[i].prev_index = -1;
		for(int j=0; j<MAX_EVENT; j++) {
			if (event[i].prev == &event[j]) {
				vm_state.event[i].prev_index = j;
				break;
			}
		}
	}
	vm_state.first_free_event_index = -1;
	for(int j=0; j<MAX_EVENT; j++) {
		if (first_free_event == &event[j]) {
			vm_state.first_free_event_index = j;
			break;
		}
	}
	vm_state.first_fire_event_index = -1;
	for(int j=0; j<MAX_EVENT; j++) {
		if (first_fire_event == &event[j]) {
			vm_state.first_fire_event_index = j;
			break;
		}
	}

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void EVENT::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// restore events
	event_clocks = vm_state.event_clocks;

	for(int i=0; i<MAX_EVENT; i++) {
		event[i].event_id = vm_state.event[i].event_id;
		event[i].expired_clock = vm_state.event[i].expired_clock;
		event[i].loop_clock = vm_state.event[i].loop_clock;
		event[i].active = vm_state.event[i].active ? true : false;
		event[i].index = vm_state.event[i].index;

		if (vm_state.event[i].class_name[0] == '\0') {
			event[i].device = NULL;
		} else {
			event[i].device = vm->get_device(vm_state.event[i].class_name, vm_state.event[i].identifier);
		}
		if (vm_state.event[i].next_index < 0) {
			event[i].next = NULL;
		} else {
			event[i].next = &event[vm_state.event[i].next_index];
		}
		if (vm_state.event[i].prev_index < 0) {
			event[i].prev = NULL;
		} else {
			event[i].prev = &event[vm_state.event[i].prev_index];
		}
	}
	if (vm_state.first_free_event_index < 0) {
		first_free_event = NULL;
	} else {
		first_free_event = &event[vm_state.first_free_event_index];
	}
	if (vm_state.first_fire_event_index < 0) {
		first_fire_event = NULL;
	} else {
		first_fire_event = &event[vm_state.first_fire_event_index];
	}
}
