/** @file event.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.11.29-

	@brief [ event manager ]
*/

#ifndef _EVENT_H_
#define _EVENT_H_

#include "vm.h"
#include "../emu.h"
#include "device.h"

#define MAX_CPU		8
#define MAX_SOUND	8
#define MAX_LINES	(LINES_PER_FRAME * 8)
#define MAX_EVENT	64
#define NO_EVENT	-1
#ifdef _BML3MK5
#define MAX_DISPLAY	2
#endif

#ifdef _DEBUG_LOG
//#define _DEBUG_SOUND_LATE
#endif

/**
	@brief Event manager
*/
class EVENT : public DEVICE
{
private:
	// event manager
	typedef struct {
		DEVICE* device;
		int cpu_clocks;
		int update_clocks;
		int accum_clocks;
	} cpu_t;
	cpu_t d_cpu[MAX_CPU];
	int dcount_cpu;

	int vclocks[MAX_LINES];
	int event_power;
	int cpu_power;
	int event_remain;
	int cpu_remain, cpu_accum, cpu_done;
	uint64 event_clocks;
	int frame_split_num;

	typedef struct event_t {
		DEVICE* device;
		int event_id;
		uint64 expired_clock;
		uint32 loop_clock;
		bool active;
		int index;
		event_t *next;
		event_t *prev;
	} event_t;
	event_t event[MAX_EVENT];
	event_t *first_free_event;
	event_t *first_fire_event;

	DEVICE* frame_event[MAX_EVENT];
	DEVICE* vline_event[MAX_EVENT];
	int frame_event_count, vline_event_count;

	double frames_per_sec, next_frames_per_sec;
	int lines_per_frame, next_lines_per_frame;

	int vline_split[FRAME_SPLIT_NUM*2+1];

	void update_event(int clock);
	void insert_event(event_t *event_handle);

	// sound manager
	DEVICE* d_sound[MAX_SOUND];
	int dcount_sound;

#ifdef USE_AUDIO_U8
	uint8* sound_buffer;
	uint8* sound_empty;
#else
	uint16* sound_buffer;
	uint16* sound_empty;
#endif
	int32* sound_tmp;
	int sound_buffer_ptr;
	int sound_buffer_overflow;
	int sound_rate;
	int sound_samples;
	int sound_tmp_samples;
	int accum_samples;
	int update_samples;
//	int update_sound_rate_f;
//	int sound_rate_f;
	int master_volume;
	bool *now_rec_sound;
	void mix_sound(int samples);
	void update_sound(int v);
	void record_sound(int samples);

#ifdef _DEBUG_LOG
	bool initialize_done;
#endif
#ifdef _DEBUG_SOUND_LATE
	uint64* sound_tmp_clocks;
#endif

#ifdef _BML3MK5
	// display
	DEVICE *d_display[MAX_DISPLAY];
	int dcount_display;

	bool *vm_pause;
#endif

	// for resume
#pragma pack(1)
	struct vm_state_st {
		uint64 event_clocks;
		int first_free_event_index;
		int first_fire_event_index;

		struct event_t {
			char class_name[12];
			char identifier[4];
			int event_id;
			uint64 expired_clock;
			uint32 loop_clock;
			uint8  active;
			int index;
			int next_index;
			int prev_index;

			char reserved[3];
		} event[MAX_EVENT];
	} vm_state;
#pragma pack()

public:
	EVENT(VM* parent_vm, EMU* parent_emu, char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("EVENT");

		dcount_cpu = dcount_sound = 0;
		frame_event_count = vline_event_count = 0;

		// initialize event
		for(int i = 0; i < MAX_EVENT; i++) {
			event[i].device = NULL;
			event[i].event_id = -1;
			event[i].expired_clock = 0;
			event[i].loop_clock = 0;
			event[i].active = false;
			event[i].index = i;
			event[i].next = (i + 1 < MAX_EVENT) ? &event[i + 1] : NULL;
		}
		first_free_event = &event[0];
		first_fire_event = NULL;

		event_clocks = 0;

		// force update timing in the first frame
		frames_per_sec = 0.0;
		lines_per_frame = 0;
		next_frames_per_sec = FRAMES_PER_SEC;
		next_lines_per_frame = LINES_PER_FRAME;

#ifdef _DEBUG_LOG
		initialize_done = false;
#endif
#ifdef _BML3MK5
		dcount_display = 0;
#endif
	}
	~EVENT() {}

	// common functions
	void initialize();
	void release();
	void reset();
	void update_config();

	// common event functions
	int event_manager_id() {
		return this_device_id;
	}
	void set_frames_per_sec(double new_frames_per_sec) {
		next_frames_per_sec = new_frames_per_sec;
	}
	void set_lines_per_frame(int new_lines_per_frame) {
		next_lines_per_frame = new_lines_per_frame;
		if (next_lines_per_frame >= MAX_LINES) next_lines_per_frame = MAX_LINES;
	}
	void register_event(DEVICE* device, int event_id, double usec, bool loop, int* register_id);
	void register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id);
	void cancel_event(int register_id);
	void register_frame_event(DEVICE* device);
	void register_vline_event(DEVICE* device);
	uint64 current_clock();
	uint64 passed_clock(uint64 prev);
	uint32 get_cpu_pc(int index);

	// unique functions
	double frame_rate() {
		return next_frames_per_sec;
	}
	void drive(int split_num);

	void initialize_sound(int rate, int samples);
#ifdef USE_AUDIO_U8
	uint8* create_sound(int* extra_frames, int samples);
#else
	uint16* create_sound(int* extra_frames, int samples);
#endif
	void set_volume(int volume);

	void set_context_cpu(DEVICE* device, int clocks) {
		int index = dcount_cpu++;
		d_cpu[index].device = device;
		d_cpu[index].cpu_clocks = clocks;
		d_cpu[index].accum_clocks = 0;
	}
	void set_context_cpu(DEVICE* device) {
		set_context_cpu(device, CPU_CLOCKS);
	}
	void set_context_sound(DEVICE* device) {
		d_sound[dcount_sound++] = device;
	}
#ifdef _BML3MK5
	void set_context_display(DEVICE* device) {
		d_display[dcount_display++] = device;
	}
#endif
	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);
};

#endif

