/** @file win32_input.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 input ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "fifo.h"
#include "fileio.h"
#include "win32_input.h"

#define KEY_KEEP_FRAMES 3

void EMU::EMU_INPUT()
{
	vm_key_status = NULL;
	vm_key_status_size = 0;

	pressed_global_key = false;
}

void EMU::initialize_input()
{
	out_debug(_T("EMU::initialize_input"));

	// initialize status
	memset(key_status, 0, sizeof(key_status));
	memset(joy_status, 0, sizeof(joy_status));
	memset(mouse_status, 0, sizeof(mouse_status));

	// initialize joysticks
	use_joystick = (config.show_dlg & 4) ? true : false;
	initialize_joystick();

	// mouse emulation is disenabled
	mouse_enabled = false;

#ifdef USE_SHIFT_NUMPAD_KEY
	// initialize shift+numpad conversion
	memset(key_converted, 0, sizeof(key_converted));
	key_shift_pressed = key_shift_released = false;
#endif
#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = new FIFO(65536);
	autokey_buffer->clear();
	autokey_phase = 31;	// wait a few sec.
	autokey_shift = 0;
	autokey_enabled = false;
#endif
	lost_focus = false;
	key_mod = 0;
}

void EMU::initialize_joystick()
{
#ifdef USE_JOYSTICK
	memset(joy_enabled, 0, sizeof(joy_enabled));
	memset(joyid_map, 0xff, sizeof(joyid_map));
	reset_joystick();

	memset(&dummy_key, 0, sizeof(dummy_key));
	dummy_key.type = INPUT_KEYBOARD;
	dummy_key.ki.wVk = 0x8f;
	dummy_key.ki.wScan = MapVirtualKey(0x8f, 0);
	dummy_key.ki.dwFlags = KEYEVENTF_EXTENDEDKEY;

	send_dummy_key = 0;
#endif
}

void EMU::reset_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	int i = 0;
	joy_num = joyGetNumDevs();
	for(int joyid = 0; joyid < joy_num && joyid < 16 && i < 2; joyid++) {
		JOYINFO joyinfo;
		if(joyGetPos(joyid, &joyinfo) == JOYERR_NOERROR) {
			joyid_map[i] = joyid;

			JOYCAPS joycaps;
			joyGetDevCaps(joyid, &joycaps, sizeof(JOYCAPS));
			uint32 x = (joycaps.wXmin + joycaps.wXmax) / 2;
			joy_xmin[i] = (joycaps.wXmin + x) / 2;
			joy_xmax[i] = (joycaps.wXmax + x) / 2;
			uint32 y = (joycaps.wYmin + joycaps.wYmax) / 2;
			joy_ymin[i] = (joycaps.wYmin + y) / 2;
			joy_ymax[i] = (joycaps.wYmax + y) / 2;

			joy_enabled[i] = true;

			i++;
		}
	}
#endif
}

void EMU::release_input()
{
	// release mouse
	if(mouse_enabled) {
		disenable_mouse();
	}

#ifdef USE_AUTO_KEY
	// release autokey buffer
	if(autokey_buffer) {
		autokey_buffer->release();
		delete autokey_buffer;
	}
#endif
}

void EMU::release_joystick()
{
#ifdef USE_JOYSTICK
	// release joystick
	for(int i = 0; i < 2; i++) {
		joy_enabled[i] = false;
	}
	memset(joy_status, 0, sizeof(joy_status));
	memset(joyid_map, 0xff, sizeof(joyid_map));
#endif
}

void EMU::update_input()
{
#ifdef USE_SHIFT_NUMPAD_KEY
	// update numpad key status
	if(key_shift_pressed && !key_shift_released) {
		if(key_status[VK_SHIFT] == 0) {
			// shift key is newly pressed
			key_status[VK_SHIFT] = 0x80;
#ifdef NOTIFY_KEY_DOWN
			vm->key_down(VK_SHIFT, false);
#endif
		}
	}
	else if(!key_shift_pressed && key_shift_released) {
		if(key_status[VK_SHIFT] != 0) {
			// shift key is newly released
			key_status[VK_SHIFT] = 0;
#ifdef NOTIFY_KEY_DOWN
			vm->key_up(VK_SHIFT);
#endif
			// check l/r shift
			if(!(GetAsyncKeyState(VK_LSHIFT) & 0x8000)) key_status[VK_LSHIFT] &= 0x7f;
			if(!(GetAsyncKeyState(VK_RSHIFT) & 0x8000)) key_status[VK_RSHIFT] &= 0x7f;
		}
	}
	key_shift_pressed = key_shift_released = false;
#endif

	// release keys
#ifdef USE_AUTO_KEY
	if(lost_focus && autokey_phase == 0) {
#else
	if(lost_focus) {
#endif
		// we lost key focus so release all pressed keys
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x80) {
				key_status[i] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	else {
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x7f) {
				key_status[i] = (key_status[i] & 0x80) | ((key_status[i] & 0x7f) - 1);
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	lost_focus = false;

#ifdef USE_JOYSTICK
	update_joystick();
#endif
#ifdef USE_KEY_TO_JOY
	// emulate joystick #1 with keyboard
	if(key_status[0x26]) joy_status[0] |= 0x01;	// up
	if(key_status[0x28]) joy_status[0] |= 0x02;	// down
	if(key_status[0x25]) joy_status[0] |= 0x04;	// left
	if(key_status[0x27]) joy_status[0] |= 0x08;	// right
#endif

	// update mouse status
	memset(mouse_status, 0, sizeof(mouse_status));
	if(mouse_enabled) {
		// get current status
		POINT pt;
		GetCursorPos(&pt);
		ScreenToClient(main_window_handle, &pt);
		mouse_status[0]  = pt.x - display_width / 2;
		mouse_status[1]  = pt.y - display_height / 2;
		mouse_status[2]  = (GetAsyncKeyState(VK_LBUTTON) & 0x8000) ? 1 : 0;
		mouse_status[2] |= (GetAsyncKeyState(VK_RBUTTON) & 0x8000) ? 2 : 0;
		mouse_status[2] |= (GetAsyncKeyState(VK_MBUTTON) & 0x8000) ? 4 : 0;
		// move mouse cursor to the center of window
		if(!(mouse_status[0] == 0 && mouse_status[1] == 0)) {
			pt.x = display_width / 2;
			pt.y = display_height / 2;
			ClientToScreen(main_window_handle, &pt);
			SetCursorPos(pt.x, pt.y);
		}
	}
#ifdef USE_LIGHTPEN
	if(true) {
		// get current status
		POINT pt;
		GetCursorPos(&pt);
		ScreenToClient(main_window_handle, &pt);
		// adjust point
		pt.x = screen_width * (pt.x - screen_dest_x) / stretched_width;
		pt.y = screen_height * (pt.y - screen_dest_y) / stretched_height;

		mouse_status[0] = pt.x;
		mouse_status[1] = pt.y;
		mouse_status[2] = (GetAsyncKeyState(VK_LBUTTON) & 0x8000 ? 1 : 0) | (GetAsyncKeyState(VK_RBUTTON) & 0x8000 ? 2 : 0);
	}
#endif
}

void EMU::update_autokey()
{
#ifdef USE_AUTO_KEY
	// auto key
	switch(autokey_phase) {
	case 1:
		if(autokey_buffer && !autokey_buffer->empty()) {
#ifdef _BML3MK5
			// update graph key status
			int graph = autokey_buffer->read_not_remove(0) & AUTO_KEY_GRAPH_MASK;
			if(graph && !autokey_shift) {
#ifdef USE_AUTO_KEY_VM_CODE
				vm_key_down(AUTO_KEY_GRAPH);
#else
				key_down(AUTO_KEY_GRAPH, 0);
#endif
				out_debug(_T("autokey phase:% 2d: graph down"), autokey_phase);
				autokey_shift = graph;
				autokey_phase++;
				break;
			}
#endif
			// update shift key status
			int shift = autokey_buffer->read_not_remove(0) & AUTO_KEY_SHIFT_MASK;
			if(shift && !autokey_shift) {
#ifdef USE_AUTO_KEY_VM_CODE
				vm_key_down(AUTO_KEY_SHIFT);
#else
				key_down(AUTO_KEY_SHIFT, 0);
#endif
				out_debug(_T("autokey phase:% 2d: shift down"), autokey_phase);
				autokey_shift = shift;
				autokey_phase++;
				break;
			}
			autokey_shift = 0;
		}
		autokey_phase++;
		break;
	case 2:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_code = autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK;
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_down(autokey_code);
#else
			autokey_code = key_down(autokey_code, 0);
#endif
			out_debug(_T("autokey phase:% 2d: key down %x"), autokey_phase, autokey_code);
		}
		autokey_phase++;
		break;
	case USE_AUTO_KEY:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debug(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 2:
		if(autokey_shift & AUTO_KEY_SHIFT_MASK) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(AUTO_KEY_SHIFT);
#else
			key_up(AUTO_KEY_SHIFT);
#endif
			out_debug(_T("autokey phase:% 2d: shift up:"), autokey_phase);
		}
#ifdef _BML3MK5
		if(autokey_shift & AUTO_KEY_GRAPH_MASK) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(AUTO_KEY_GRAPH);
#else
			key_up(AUTO_KEY_GRAPH);
#endif
			out_debug(_T("autokey phase:% 2d: graph up"), autokey_phase);
		}
#endif
		autokey_shift = 0;
		if(autokey_buffer && !autokey_buffer->empty()) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK);
#else
			key_up(autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK);
#endif
		}
		out_debug(_T("autokey phase:% 2d: key up %x"), autokey_phase, autokey_code);
		autokey_code = 0;
		autokey_phase++;
		break;
	case USE_AUTO_KEY + 3:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debug(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 4:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// wait enough while vm analyzes one line
			int code = autokey_buffer->read();
			out_debug(_T("autokey phase:% 2d: code %x"), autokey_phase, code);
			if(code == AUTO_KEY_RETURN) {
				autokey_phase++;	// wait 30 frames
				break;
			}
		}
		// through
	case 30:
	case 150:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_phase = 1;
		}
		else {
			stop_auto_key();
			autokey_phase = 0;
		}
		break;
	default:
		if(autokey_phase) {
			autokey_phase++;
		}
		break;
	}
#endif
}

void EMU::recognized_key(uint8 key_code)
{
#ifdef USE_AUTO_KEY
	if(autokey_phase == 0) return;
	if(autokey_phase == USE_AUTO_KEY && autokey_code) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		if (autokey_code == key_code) {
			autokey_phase = USE_AUTO_KEY + 1;
		}
	}
	if(autokey_phase == USE_AUTO_KEY + 3 && key_code == AUTO_KEY_NONE) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		autokey_phase =  USE_AUTO_KEY + 4;
	}
#endif
}

void EMU::update_joystick()
{
#ifdef USE_JOYSTICK
	memset(joy_status, 0, sizeof(joy_status));

	if (!use_joystick) return;

	// update joystick status
	for(int i = 0; i < 2; i++) {
		JOYINFOEX joyinfoex;
		joyinfoex.dwSize = sizeof(JOYINFOEX);
		joyinfoex.dwFlags = JOY_RETURNALL;
		if(joy_enabled[i] == true && joyid_map[i] != -1) {
			if (joyGetPosEx(joyid_map[i], &joyinfoex) == JOYERR_NOERROR) {
				if(joyinfoex.dwYpos < joy_ymin[i]) joy_status[i] |= 0x01;
				if(joyinfoex.dwYpos > joy_ymax[i]) joy_status[i] |= 0x02;
				if(joyinfoex.dwXpos < joy_xmin[i]) joy_status[i] |= 0x04;
				if(joyinfoex.dwXpos > joy_xmax[i]) joy_status[i] |= 0x08;
				joy_status[i] |= ((joyinfoex.dwButtons & 0x0fffffff) << 4);
			} else {
				joy_enabled[i] = false;
			}
		}
		if (joy_status[i] && send_dummy_key <= 0) {
			// send dummy key event, because delay start time of screen saver.
			SendInput(1, &dummy_key, sizeof(INPUT));
			send_dummy_key = 60;
		} else {
			send_dummy_key -= (send_dummy_key > 0) ? 1 : 0;
		}
	}
#endif
}

//void EMU::key_down(int code, bool repeat)
int EMU::key_down(int code, long status)
{
	int  orig_code = code;
	bool keep_frames = false;
	bool repeat = ((HIWORD(status) & 0x4000) != 0);
	int  scan_code = (status & 0x1ff0000) >> 16;

	if(code == VK_SHIFT) {
		if(GetAsyncKeyState(VK_LSHIFT) & 0x8000) key_status[VK_LSHIFT] = 0x80;
		if(GetAsyncKeyState(VK_RSHIFT) & 0x8000) key_status[VK_RSHIFT] = 0x80;
		if(!(key_status[VK_LSHIFT] || key_status[VK_RSHIFT])) key_status[VK_LSHIFT] = 0x80;
	}
	else if(code == VK_CONTROL) {
		if(GetAsyncKeyState(VK_LCONTROL) & 0x8000) key_status[VK_LCONTROL] = 0x80;
		if(GetAsyncKeyState(VK_RCONTROL) & 0x8000) key_status[VK_RCONTROL] = 0x80;
		if(!(key_status[VK_LCONTROL] || key_status[VK_RCONTROL])) key_status[VK_LCONTROL] = 0x80;
	}
	else if(code == VK_MENU) {
		if(GetAsyncKeyState(VK_LMENU) & 0x8000) key_status[VK_LMENU] = 0x80;
		if(GetAsyncKeyState(VK_RMENU) & 0x8000) key_status[VK_RMENU] = 0x80;
		if(!(key_status[VK_LMENU] || key_status[VK_RMENU])) key_status[VK_LMENU] = 0x80;
	}
	else if(code == 0xf0) {
		code = VK_CAPITAL;
		keep_frames = true;
	}
#ifdef _BML3MK5
	else if(code == 0xf1) {
		code = VK_NONCONVERT;
//		keep_frames = true;
	}
#endif
	else if(code == 0xf2) {
		code = VK_KANA;
		keep_frames = true;
	}
	else if(code == 0xf3 || code == 0xf4) {
		code = VK_KANJI;
//		keep_frames = true;
	}

#ifdef _BML3MK5
	// convert numpad keys
	if (scan_code >= 0x47 && scan_code <= 0x53) {
		if (scancode2vkey[scan_code - 0x47] != 0) code = scancode2vkey[scan_code - 0x47];
	}
#endif

#ifdef USE_SHIFT_NUMPAD_KEY
	if(code == VK_SHIFT) {
		key_shift_pressed = true;
		return;
	}
	else if(numpad_table[code] != 0) {
		if(key_shift_pressed || key_shift_released) {
			key_converted[code] = 1;
			key_shift_pressed = true;
			code = numpad_table[code];
		}
	}
#endif
#ifdef DONT_KEEEP_KEY_PRESSED
		if(!(code == VK_SHIFT || code == VK_CONTROL || code == VK_MENU)) {
			key_status[code] = KEY_KEEP_FRAMES;
		}
		else
#endif
	key_status[code] = keep_frames ? (KEY_KEEP_FRAMES * FRAME_SPLIT_NUM) : 0x80;
#ifdef NOTIFY_KEY_DOWN
	if(keep_frames) {
		repeat = false;
	}
	vm->key_down(code, repeat);
#endif
//	out_debug("keydown: %02x->%02x s:%08x repeat:%s",orig_code,code,status,repeat ? "true" : "false");
	return code;
}

//void EMU::key_up(int code)
void EMU::key_up(int code, long status)
{
	int  orig_code = code;
	int  scan_code = (status & 0x1ff0000) >> 16;

	if(code == VK_SHIFT) {
#ifndef USE_SHIFT_NUMPAD_KEY
		if(!(GetAsyncKeyState(VK_LSHIFT) & 0x8000)) key_status[VK_LSHIFT] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RSHIFT) & 0x8000)) key_status[VK_RSHIFT] &= 0x7f;
#endif
	}
	else if(code == VK_CONTROL) {
		if(!(GetAsyncKeyState(VK_LCONTROL) & 0x8000)) key_status[VK_LCONTROL] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RCONTROL) & 0x8000)) key_status[VK_RCONTROL] &= 0x7f;
	}
	else if(code == VK_MENU) {
		if(!(GetAsyncKeyState(VK_LMENU) & 0x8000)) key_status[VK_LMENU] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RMENU) & 0x8000)) key_status[VK_RMENU] &= 0x7f;
	}
	else if(code == 0xf0) {
		code = VK_CAPITAL;
	}
#ifdef _BML3MK5
	else if(code == 0xf1) {
		code = VK_NONCONVERT;
	}
#endif
	else if(code == 0xf2) {
		code = VK_KANA;
	}
	else if(code == 0xf3 || code == 0xf4) {
		code = VK_KANJI;
	}

#ifdef _BML3MK5
	// convert numpad keys
	if (scan_code >= 0x47 && scan_code <= 0x53) {
		if (scancode2vkey[scan_code - 0x47] != 0) code = scancode2vkey[scan_code - 0x47];
	}
#endif

#ifdef USE_SHIFT_NUMPAD_KEY
	if(code == VK_SHIFT) {
		key_shift_pressed = false;
		key_shift_released = true;
		return;
	}
	else if(key_converted[code] != 0) {
		key_converted[code] = 0;
		code = numpad_table[code];
	}
#endif
	if(key_status[code]) {
		key_status[code] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
		if(!key_status[code]) {
			vm->key_up(code);
		}
#endif
	}
//	out_debug("keyup: %02x->%02x s:%08x",orig_code,code,status);
}

int EMU::vm_key_down(int code)
{
#ifdef _BML3MK5
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] = 1;
	}
	return code;
}
void EMU::vm_key_up(int code)
{
#ifdef _BML3MK5
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] = 0;
	}
}

#ifdef USE_BUTTON
void EMU::press_button(int num)
{
	int code = buttons[num].code;

	if(code) {
		key_down(code, false);
		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
	}
	else {
		// code=0: reset virtual machine
		vm->reset();
	}
}
#endif

void EMU::enable_mouse()
{
	// enable mouse emulation
	if(!mouse_enabled) {
		// hide mouse cursor
		ShowCursor(FALSE);
		// move mouse cursor to the center of window
		POINT pt;
		pt.x = display_width / 2;
		pt.y = display_height / 2;
		ClientToScreen(main_window_handle, &pt);
		SetCursorPos(pt.x, pt.y);
	}
	mouse_enabled = true;
}

void EMU::disenable_mouse()
{
	// disenable mouse emulation
	if(mouse_enabled) {
		ShowCursor(TRUE);
	}
	mouse_enabled = false;
}

void EMU::toggle_mouse()
{
	// toggle mouse enable / disenable
	if(mouse_enabled) {
		disenable_mouse();
	}
	else {
		enable_mouse();
	}
}

void EMU::enable_joystick()
{
	// reinitialize
	initialize_joystick();
}

#ifdef USE_AUTO_KEY
bool EMU::start_auto_key()
{
	bool rc = true;

	stop_auto_key();

	autokey_code = 0;
	if(OpenClipboard(NULL)) {
		HANDLE hClip = GetClipboardData(CF_TEXT);
		if(hClip) {
			autokey_buffer->clear();
			int prev_code = 0;
			char* buf = (char*)GlobalLock(hClip);
			int size = (int)strlen(buf), prev_kana = 0;
			parse_auto_key(buf, size, prev_code, prev_kana);
			if(prev_kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
				if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
			}
			GlobalUnlock(hClip);

			autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
			autokey_shift = 0;
			autokey_enabled = true;
		}
		CloseClipboard();
	} else {
		rc = false;
	}

	update_config();

	return rc;
}

bool EMU::start_auto_key(_TCHAR *file_path)
{
	FILEIO *fp;
	char buf[1025];
	bool rc = true;

	stop_auto_key();

	autokey_code = 0;
	fp = new FILEIO();
	int prev_kana = 0;
	if (fp->Fopen(file_path, FILEIO_READ_BINARY)) {
		autokey_buffer->clear();
		int prev_code = 0;
		int size = 0;
		while((size = (int)fp->Fread(buf, sizeof(char), 1024)) > 0) {
			parse_auto_key(buf, size, prev_code, prev_kana);
		}
		fp->Fclose();

		if(prev_kana) {
			autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
			if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
		}

		autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
		autokey_shift = 0;
		autokey_enabled = true;
		_tcsncpy(config.opened_autokey_path, file_path, _MAX_PATH);
	} else {
		out_log(LOG_ERROR, _T("Auto key file couldn't be opened."));
		rc = false;
	}
	delete fp;

	copy_initial_path(config.initial_autokey_path, file_path);

	update_config();

	return rc;
}

void EMU::parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana)
{
	for(int i = 0; i < size; i++) {
		int code = buf[i] & 0xff;
#ifdef _BML3MK5
		if(prev_code == 0xd && code == 0xa) {
			prev_code = code;
			continue;	// cr-lf
		}
#else
		if((0x81 <= code && code <= 0x9f) || 0xe0 <= code) {
			i++;	// kanji ?
			continue;
		}
		else if(code == 0xa) {
			continue;	// cr-lf
		}
#endif
		if((code = autokey_table[code]) != 0) {
#ifdef _BML3MK5
			int kana = code & 0x3000;
			if(kana - prev_kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(kana - prev_kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
			if(prev_kana - kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(prev_kana - kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#else
			int kana = code & 0x200;
			if(prev_kana != kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#endif
			prev_kana = kana;
#if defined(USE_AUTO_KEY_NO_CAPS)
			if((code & AUTO_KEY_SHIFT_MASK) && !(code & (AUTO_KEY_UPPER_MASK | AUTO_KEY_LOWER_MASK))) {
#elif defined(USE_AUTO_KEY_CAPS)
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_LOWER_MASK)) {
#else
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_UPPER_MASK)) {
#endif
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_SHIFT_MASK);
			}
#ifdef _BML3MK5
			else if (code & AUTO_KEY_GRAPH_MASK) {
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_GRAPH_MASK);
			}
#endif
			else {
				autokey_buffer->write(code & AUTO_KEY_MASK);
			}
		}
		prev_code =code;
	}
}

void EMU::stop_auto_key()
{
#ifndef _BML3MK5
	if(autokey_shift) {
		key_up(AUTO_KEY_SHIFT);
	}
#else
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_SHIFT);
	}
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_GRAPH);
	}
	if (vm_key_status) {
		memset(vm_key_status, 0, sizeof(uint8) * vm_key_status_size);
	}
#endif
	autokey_phase = (autokey_phase > 30 ? autokey_phase : 0);
	autokey_shift = 0;
	autokey_enabled = false;

	memset(config.opened_autokey_path, 0, sizeof(config.opened_autokey_path));

	update_config();
}
#endif
