/** @file win32_socket.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.27 -

	@brief [ win32 socket ]
*/

#include "emu.h"
#include "vm/vm.h"

void EMU::initialize_socket()
{
	// init winsock
	WSADATA wsaData;
	WSAStartup(0x0101, &wsaData);

	// init sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
		soc[i] = INVALID_SOCKET;
		socket_delay[i] = 0;
		recv_buffer[i] = NULL;
		recv_r_ptr[i] = recv_w_ptr[i] = 0;
	}
}

void EMU::release_socket()
{
	// release sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
		if(soc[i] != INVALID_SOCKET) {
			shutdown(soc[i], 2);
			closesocket(soc[i]);
		}
		delete[] recv_buffer[i];
	}

	// release winsock
	WSACleanup();
}

void EMU::socket_connected(int ch)
{
	// winmain notify that network is connected
	out_debug(_T("connected. ch=%d"), ch);
	vm->network_connected(ch);
}

void EMU::socket_disconnected(int ch)
{
	// winmain notify that network is disconnected
	if(!socket_delay[ch]) {
		socket_delay[ch] = 1;//56;
	}
	out_debug(_T("disconnected. ch=%d"), ch);
}

void EMU::socket_writeable(int ch)
{
	// winmain notify that network is writeable
	out_debug(_T("writeable. ch=%d"), ch);
	vm->network_writeable(ch);
}

void EMU::socket_readable(int ch)
{
	// winmain notify that network is readable
	out_debug(_T("readable. ch=%d"), ch);
	vm->network_readable(ch);
}

void EMU::update_socket()
{
	for(int i = 0; i < SOCKET_MAX; i++) {
		if(recv_r_ptr[i] < recv_w_ptr[i]) {
			// get buffer
			int size0, size1;
			uint8* buf0 = vm->get_recvbuffer0(i, &size0, &size1);
			uint8* buf1 = vm->get_recvbuffer1(i);

			int size = recv_w_ptr[i] - recv_r_ptr[i];
			if(size > size0 + size1) {
				size = size0 + size1;
			}
			char* src = &recv_buffer[i][recv_r_ptr[i]];
			recv_r_ptr[i] += size;

			if(size <= size0) {
				memcpy(buf0, src, size);
			}
			else {
				memcpy(buf0, src, size0);
				memcpy(buf1, src + size0, size - size0);
			}
			vm->inc_recvbuffer_ptr(i, size);
		}
		else if(socket_delay[i] != 0) {
			if(--socket_delay[i] == 0) {
				vm->network_disconnected(i);
			}
		}
	}
}

void EMU::socket_accept(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	int new_ch = -1;

	SOCKET new_soc;
	struct sockaddr_in new_sockaddr;
	int new_socklen;

	new_socklen = sizeof(new_sockaddr);
	memset(&new_sockaddr, 0, sizeof(new_sockaddr));

	if ((new_soc = accept(soc[ch], (struct sockaddr *)&new_sockaddr, &new_socklen)) == INVALID_SOCKET) {
		int ecode = WSAGetLastError();
		out_log(LOG_ERROR, _T("accept error. ch=%d code=%d"), ch, ecode);
		return;
	}

	// find empty socket
	new_ch = get_socket_channel();
	if (new_ch < 0) {
		closesocket(new_soc);
		return;
	}

	// note that i'll fire the writeable event only, won't fire the connected event.
	if(WSAAsyncSelect(new_soc, main_window_handle, WM_SOCKET0 + new_ch, FD_WRITE | FD_READ | FD_CLOSE) == SOCKET_ERROR) {
		closesocket(new_soc);
		int ecode = WSAGetLastError();
		out_log(LOG_ERROR, _T("EMU::socket_accept: WSAAsyncSelect error. new_ch=%d code=%d"), new_ch, ecode);
		return;
	}
	out_debug(_T("accepted from 0x%08x"), new_sockaddr.sin_addr.s_addr);

	soc[new_ch] = new_soc;
	is_tcp[new_ch] = is_tcp[ch];

	vm->network_accepted(ch, new_ch);

	out_debug(_T("EMU::socket_accept: ok. ch=%d is_tcp=%s new_ch=%d is_tcp=%s"), ch, is_tcp[ch] ? _T("true") : _T("false"), new_ch, is_tcp[new_ch] ? _T("true") : _T("false"));
	return;
}

bool EMU::init_socket_tcp(int ch, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		out_log(LOG_ERROR, _T("EMU::init_socket_tcp: invalid socket channel. ch=%d"), ch);
		return false;
	}

	is_tcp[ch] = true;

	if(soc[ch] != INVALID_SOCKET) {
		disconnect_socket(ch);
	}
	if((soc[ch] = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
		int ecode = WSAGetLastError();
		out_log(LOG_ERROR, _T("EMU::init_socket_tcp: socket error. ch=%d code=%d"), ch, ecode);
		return false;
	}
	if (!server) {
		if(WSAAsyncSelect(soc[ch], main_window_handle, WM_SOCKET0 + ch, FD_CONNECT | FD_WRITE | FD_READ | FD_CLOSE) == SOCKET_ERROR) {
			closesocket(soc[ch]);
			int ecode = WSAGetLastError();
			out_log(LOG_ERROR, _T("EMU::init_socket_tcp: WSAAsyncSelect error. ch=%d code=%d"), ch, ecode);
			soc[ch] = INVALID_SOCKET;
			return false;
		}
	} else {
		if(WSAAsyncSelect(soc[ch], main_window_handle, WM_SOCKET0 + ch, FD_ACCEPT | FD_CLOSE) == SOCKET_ERROR) {
			closesocket(soc[ch]);
			int ecode = WSAGetLastError();
			out_log(LOG_ERROR, _T("EMU::init_socket_tcp: WSAAsyncSelect error. ch=%d code=%d"), ch, ecode);
			soc[ch] = INVALID_SOCKET;
			return false;
		}
	}
	recv_r_ptr[ch] = recv_w_ptr[ch] = 0;
	return true;
}

bool EMU::init_socket_udp(int ch, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		out_log(LOG_ERROR, _T("EMU::init_socket_udp: invalid socket channel. ch=%d"), ch);
		return false;
	}

	is_tcp[ch] = false;

	disconnect_socket(ch);
	if((soc[ch] = socket(PF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
		int ecode = WSAGetLastError();
		out_log(LOG_ERROR, _T("EMU::init_socket_udp: socket error. ch=%d code=%d"), ch, ecode);
		return false;
	}
	if(WSAAsyncSelect(soc[ch], main_window_handle, WM_SOCKET0 + ch, FD_CONNECT | FD_WRITE | FD_READ | FD_CLOSE) == SOCKET_ERROR) {
		closesocket(soc[ch]);
		int ecode = WSAGetLastError();
		out_log(LOG_ERROR, _T("EMU::init_socket_udp: WSAAsyncSelect error. ch=%d code=%d"), ch, ecode);
		soc[ch] = INVALID_SOCKET;
		return false;
	}
	recv_r_ptr[ch] = recv_w_ptr[ch] = 0;
	return true;
}

bool EMU::connect_socket(int ch, uint32 ipaddr, int port, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		out_log(LOG_ERROR, _T("EMU::connect_socket: invalid socket channel. ch=%d"), ch);
		return false;
	}

	struct sockaddr_in tcpaddr;
	tcpaddr.sin_family = AF_INET;
	tcpaddr.sin_addr.s_addr = ipaddr;
	tcpaddr.sin_port = htons((unsigned short)port);
	memset(tcpaddr.sin_zero, (int)0, sizeof(tcpaddr.sin_zero));

	if (!server) {
		if (ipaddr == INADDR_ANY || ipaddr == INADDR_NONE) {
			// client cannot use
			out_log(LOG_ERROR, _T("Cannot create socket as the server."));
			return false;
		}
		if(connect(soc[ch], (struct sockaddr *)&tcpaddr, sizeof(tcpaddr)) == SOCKET_ERROR) {
			int ecode = WSAGetLastError();
			if(ecode != WSAEWOULDBLOCK) {
				out_log(LOG_ERROR, _T("connect error. ch=%d code=%d"), ch, ecode);
				return false;
			}
			out_debug(_T("EMU::connect_socket: connect would block. ch=%d code=%d"), ch, ecode);
		}
	} else {
		if (bind(soc[ch], (struct sockaddr *)&tcpaddr, sizeof(tcpaddr)) == SOCKET_ERROR) {
			closesocket(soc[ch]);
			int ecode = WSAGetLastError();
			out_log(LOG_ERROR, _T("bind error. ch=%d code=%d"), ch, ecode);
			soc[ch] = INVALID_SOCKET;
			return false;
		}
		if (is_tcp[ch]) {
			if (listen(soc[ch], 1) == SOCKET_ERROR) {
				closesocket(soc[ch]);
				int ecode = WSAGetLastError();
				out_log(LOG_ERROR, _T("listen error. ch=%d code=%d"), ch, ecode);
				soc[ch] = INVALID_SOCKET;
				return false;
			}
		}
	}
	return true;
}

bool EMU::connect_socket(int ch, const _TCHAR *hostname, int port, bool server)
{
	// IPv4 only
	uint32 ipaddr;

	if (!get_ipaddr(hostname, port, &ipaddr)) {
		return false;
	}

	return connect_socket(ch, ipaddr, port, server);
}

bool EMU::get_ipaddr(const _TCHAR *hostname, int port, uint32 *ipaddr)
{
	HOSTENT *ent;
#ifdef _UNICODE
	char host[_MAX_PATH];
	wcstombs(host, hostname, _MAX_PATH);
#else
	const char *host = hostname;
#endif
	ent = gethostbyname((const char *)host);
	if (ent == NULL || ent->h_addr_list[0] == NULL) {
		// unresolved hostname to ipaddr
		out_log(LOG_ERROR, _T("resolve error: hostname=%s"), host);
		return false;
	}
	// IPv4 only
	*ipaddr = *(uint32 *)ent->h_addr_list[0];

	return true;
}

void EMU::disconnect_socket(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	if(soc[ch] != INVALID_SOCKET) {
		shutdown(soc[ch], 2);
		closesocket(soc[ch]);
		soc[ch] = INVALID_SOCKET;

		vm->network_disconnected(ch);
	}
}

int EMU::get_socket_channel()
{
	int new_ch = -1;
	for(int i = 0; i < SOCKET_MAX; i++) {
		if (soc[i] == INVALID_SOCKET) {
			new_ch = i;
			break;
		}
	}
	if (new_ch < 0) {
		out_log(LOG_ERROR, _T("Socket is full."));
	}
	return new_ch;
}

bool EMU::listen_socket(int ch)
{
	return false;
}

void EMU::send_data_tcp(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	if(is_tcp[ch]) {
		send_data(ch);
	}
}

void EMU::send_data_udp(int ch, uint32 ipaddr, int port)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	if(!is_tcp[ch]) {
		udpaddr[ch].sin_family = AF_INET;
		udpaddr[ch].sin_addr.s_addr = ipaddr;
		udpaddr[ch].sin_port = htons((unsigned short)port);
		memset(udpaddr[ch].sin_zero, (int)0, sizeof(udpaddr[ch].sin_zero));

		send_data(ch);
	}
}

void EMU::send_data_udp(int ch, const _TCHAR *hostname, int port)
{
	uint32 ipaddr;

	if (!get_ipaddr(hostname, port, &ipaddr)) {
		return;
	}

	send_data_udp(ch, ipaddr, port);
}

void EMU::send_data(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	// loop while send buffer is not empty or not WSAEWOULDBLOCK
	while(1) {
		// get send buffer and data size
		int size;
		uint8* buf = vm->get_sendbuffer(ch, &size);

		if(!size) {
			return;
		}
		if(is_tcp[ch]) {
			if((size = send(soc[ch], (char *)buf, size, 0)) == SOCKET_ERROR) {
				// if WSAEWOULDBLOCK, WM_SOCKET* and FD_WRITE will come later
				if(WSAGetLastError() != WSAEWOULDBLOCK) {
					disconnect_socket(ch);
					socket_disconnected(ch);
				}
				return;
			}
		}
		else {
			if((size = sendto(soc[ch], (char *)buf, size, 0, (struct sockaddr *)&udpaddr[ch], sizeof(udpaddr[ch]))) == SOCKET_ERROR) {
				// if WSAEWOULDBLOCK, WM_SOCKET* and FD_WRITE will come later
				if(WSAGetLastError() != WSAEWOULDBLOCK) {
					disconnect_socket(ch);
					socket_disconnected(ch);
				}
				return;
			}
		}
		vm->inc_sendbuffer_ptr(ch, size);
	}
}

void EMU::recv_data(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	if (recv_buffer[ch] == NULL) {
		recv_buffer[ch] = new char[SOCKET_BUFFER_MAX];
	}

	if(is_tcp[ch]) {
		int size = SOCKET_BUFFER_MAX - recv_w_ptr[ch];
		char* buf = &recv_buffer[ch][recv_w_ptr[ch]];
		if((size = recv(soc[ch], buf, size, 0)) == SOCKET_ERROR) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
		recv_w_ptr[ch] += size;
	}
	else {
		SOCKADDR_IN addr;
		int len = sizeof(addr);
		int size = SOCKET_BUFFER_MAX - recv_w_ptr[ch];
		char* buf = &recv_buffer[ch][recv_w_ptr[ch]];

		if(size < 8) {
			return;
		}
		if((size = recvfrom(soc[ch], buf + 8, size - 8, 0, (struct sockaddr *)&addr, &len)) == SOCKET_ERROR) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
		size += 8;
		buf[0] = size >> 8;
		buf[1] = size;
		buf[2] = (char)addr.sin_addr.s_addr;
		buf[3] = (char)(addr.sin_addr.s_addr >> 8);
		buf[4] = (char)(addr.sin_addr.s_addr >> 16);
		buf[5] = (char)(addr.sin_addr.s_addr >> 24);
		buf[6] = (char)addr.sin_port;
		buf[7] = (char)(addr.sin_port >> 8);
		recv_w_ptr[ch] += size;
	}
}

